module oc_mpi_combine_mod
  !!=======================
  !! Paul Sava    (C) 2000
  !! paul@sep.stanford.edu
  !! oclib v2.0
  !!=======================
  use oc_global_mod
  use oc_file_mod
  use oc_filealgebra_mod
  use oc_scale_mod
  use oc_combine_mod
  implicit none

  interface oc_mpi_column
     module procedure oc_mpi_column11          !! [y1 y2]'=[A1       e.A2]' x
     module procedure oc_mpi_column20          !! [y1 y2]'=[A1.B1    e.I ]' x
     module procedure oc_mpi_column21          !! [y1 y2]'=[A1.B1    e.A2]' x
     module procedure oc_mpi_column30          !! [y1 y2]'=[A1.B1.C1 e.I ]' x
  end interface

contains
  !----------------------------------------------------------------
  subroutine oc_mpi_column11(impi,A1,eps,A2, adj,add, x_,y1_,y2_, maxmem, op1,op2,op3,op4,op5,op6,op7,op8,op9)
    integer,            intent(in) :: impi
    real,               intent(in) :: eps
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: x_,y1_,y2_
    integer,            optional   :: maxmem
    integer, external,  optional   :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    character(len=128)             :: t_, name
    interface!!----------------------------------------------------------------
       integer function A1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function A1
       integer function A2( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function A2
    end interface!!------------------------------------------------------------
    !!  [ L ]
    !!  [eA ]x

    if(.not.present(maxmem)) maxmem=0
    call oc_scale_init(eps,maxmem)

    name='_t.chain'; if(impi==0) t_=oc_clone(T, y2_,name,maxmem)

    if (adj) then
       call oc_chain1(         A1, T,add, x_,y1_,    op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x =  A1' y1
       call oc_chain2(oc_scale,A2, T,  T, x_,y2_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x+=e A2' y2
    else
       call oc_chain1(         A1, F,add, x_,y1_,    op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y1=  A1 x
       call oc_chain2(oc_scale,A2, F,add, x_,y2_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y2=e A2 x
    end if
  end subroutine oc_mpi_column11
  !----------------------------------------------------------------
  subroutine oc_mpi_column20(impi,A1,B1,eps, adj,add, x_,y1_,y2_, maxmem, op1,op2,op3,op4,op5,op6,op7,op8,op9)
    integer,            intent(in) :: impi
    real,               intent(in) :: eps
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: x_,y1_,y2_
    integer,            optional   :: maxmem
    integer, external,  optional   :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    character(len=128)             :: t_,name
    interface!!----------------------------------------------------------------
       integer function A1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function A1
       integer function B1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function B1
    end interface!!------------------------------------------------------------
    !! [ LP]
    !! [eI ]

    if(.not.present(maxmem)) maxmem=0
    call oc_scale_init(eps,maxmem)

if(impi==0) write(0,*) "clone x_ in _t.chain"
    name='_t.chain'; if(impi==0) t_=oc_clone(F, x_,name,maxmem)

    if (adj) then
       call oc_chain2(A1,      B1, T,add, x_,y1_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x =  B1' A1' y1
       call oc_chain1(   oc_scale, T,  T, x_,y2_,    op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x+=e         y2
    else
       call oc_chain2(A1,      B1, F,add, x_,y1_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y1=  A1 B1 x
       call oc_chain1(   oc_scale, F,add, x_,y2_,    op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y2=e       x
    end if
  end subroutine oc_mpi_column20
  !----------------------------------------------------------------
  subroutine oc_mpi_column21(impi,A1,B1,eps,A2, adj,add, x_,y1_,y2_, maxmem, op1,op2,op3,op4,op5,op6,op7,op8,op9)
    integer,            intent(in) :: impi
    real,               intent(in) :: eps
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: x_,y1_,y2_
    integer,            optional   :: maxmem
    integer, external,  optional   :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    character(len=128)             :: t_, name
    interface!!----------------------------------------------------------------
       integer function A1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function A1
       integer function B1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function B1
       integer function A2( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function A2
    end interface!!------------------------------------------------------------
    !! [WL ]
    !! [eA ]

    if(.not.present(maxmem)) maxmem=0
    call oc_scale_init(eps,maxmem)
    name='_t.chain'

    if (adj) then
       if(impi==0) t_=oc_clone(T, y1_,name,maxmem)
       call oc_chain2(      A1,B1, T,add, x_,y1_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x =  B1' A1' y1
       if(impi==0) t_=oc_clone(T, y2_,name,maxmem)
       call oc_chain2(oc_scale,A2, T,  T, x_,y2_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x+=e     A2' y2
    else
       if(impi==0) t_=oc_clone(T, y1_,name,maxmem)
       call oc_chain2(      A1,B1, F,add, x_,y1_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y1=  A1 B1 x
       if(impi==0) t_=oc_clone(T, y2_,name,maxmem)
       call oc_chain2(oc_scale,A2, F,add, x_,y2_,t_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y2=e A2    x
    end if
  end subroutine oc_mpi_column21
  !----------------------------------------------------------------
  subroutine oc_mpi_column30(impi,A1,B1,C1,eps, adj,add, x_,y1_,y2_, maxmem, op1,op2,op3,op4,op5,op6,op7,op8,op9)
    integer,            intent(in) :: impi
    real,               intent(in) :: eps
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: x_,y1_,y2_
    integer,            optional   :: maxmem
    integer, external,  optional   :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    character(len=128)             :: t1_,t2_,name
    interface!!----------------------------------------------------------------
       integer function A1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function A1
       integer function B1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function B1
       integer function C1( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function C1
    end interface!!------------------------------------------------------------
    !! [WLP]
    !! [eI ]

    if(.not.present(maxmem)) maxmem=0
    call oc_scale_init(eps,maxmem)
    name='_t1.chain'; if(impi==0) t1_=oc_clone(T,  x_,name,maxmem)
    name='_t2.chain'; if(impi==0) t2_=oc_clone(T, y1_,name,maxmem)

    if (adj) then
       call oc_chain3(A1,B1,      C1, T,add, x_,y1_,t1_,t2_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x =  C1' B1' A1' y1
       call oc_chain1(      oc_scale, T,  T, x_,y2_,         op1,op2,op3,op4,op5,op6,op7,op8,op9) !! x+=e             y2
    else
       call oc_chain3(A1,B1,      C1, F,add, x_,y1_,t1_,t2_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y1=  A1 B1 C1 x
       call oc_chain1(      oc_scale, F,add, x_,y2_,         op1,op2,op3,op4,op5,op6,op7,op8,op9) !! y2=e          x
    end if
  end subroutine oc_mpi_column30
  !----------------------------------------------------------------
end module oc_mpi_combine_mod









