module oc_mpi_solverreg_mod
  !!=======================
  !! Paul Sava    (C) 2000
  !! paul@sep.stanford.edu
  !! oclib v2.0
  !!=======================
  use oc_global_mod
  use oc_tags_mod
  use oc_utils_mod
  use oc_file_mod
  use oc_scale_mod
  use oc_filealgebra_mod
  implicit none

  integer, private :: ocniter,ocmaxmem
  real,    private :: oceps
  logical, private :: ocverb, ocmmovie, ocdmovie
  integer, private :: ocimpi
contains
  !! L = linear         operator
  !! A = regularization operator
  !! W = weighting      operator
  !! S = step

  !----------------------------------------------------------------
  subroutine oc_mpi_solverreg_init(impi,niter,eps,maxmem,verb,mmovie,dmovie)
    integer           :: impi
    integer, optional :: niter, maxmem
    real,    optional :: eps
    logical, optional :: verb, mmovie, dmovie

    ocimpi = impi

    ocniter=10;     if(present(niter))     ocniter=niter
    ocmaxmem=0;     if(present(maxmem))   ocmaxmem=maxmem
    ocverb=F;       if(present(verb))       ocverb=verb
    ocmmovie=F;     if(present(mmovie))   ocmmovie=mmovie
    ocdmovie=F;     if(present(dmovie))   ocdmovie=dmovie
    oceps=0.01;     if(present(eps))         oceps=eps

    if(ocimpi==0) then
       call ocline()
       write(0,*) "niter=",niter
       write(0,*) "  eps=",eps
       call ocline()
    end if
  end subroutine oc_mpi_solverreg_init
  !----------------------------------------------------------------
  subroutine oc_mpi_solverreg( L,A,S, x_,yy_, nreg, W ,k_,x0_,res_,op1,op2,op3,op4,op5,op6,op7,op8,op9)
    optional                                 :: W, k_,x0_,res_
    character(len=128),    intent(in)        ::    k_,x0_,res_
    character(len=128),    intent(in)        :: x_, yy_
    integer, external,     optional          :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    integer, dimension(:), intent(in)        :: nreg

    type(fileinfo)                           :: file
    character(len=128)                       ::    rd_,rm_,rt_ !! residual
    character(len=128)                       :: g_,gd_,gm_ !! gradient and conjugate gradient
    character(len=128)                       :: s_,sd_,sm_ !! previous step
    character(len=128)                       ::mm_,md_     !! model/data movie
    character(len=128)                       :: name
    character(len=128)                       ::    tm_,td_

    integer                                  :: i, stat
    logical                                  :: forget
    real, dimension(1)                       :: res

    interface
       integer function L( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function L
       integer function A( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function A
       integer function W( adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(st)
         integer, external, optional :: op1,op2,op3,op4,op5,op6,op7,op8,op9
         logical, intent (in)        :: adj, add
         character(len=128)          :: x_, yy_
       end function W
       integer function S( forget,x_,g_,rr_,gg_,s_,ss_,maxmem) result(st)
         integer, optional                :: maxmem
         logical                          :: forget
         character(len=128)               ::  x_,  g_,  s_
         character(len=128), dimension(:) :: rr_, gg_, ss_
       end function S
    end interface

    if(ocimpi==0) then
       call oc_allocatefile(file,x_,ocmaxmem)

       name="RS_Rm";    rm_ = oc_allocate(F, name, file%esize, nreg) !! model residual
       name="RS_Rd";    rd_ = oc_clone(F, yy_,name, ocmaxmem) !! data  residual

       name="RS_g";      g_ = oc_clone(F,  x_,name, ocmaxmem)
       name="RS_Gm";    gm_ = oc_allocate(F, name, file%esize, nreg) !! model conj grad
       name="RS_Gd";    gd_ = oc_clone(F, yy_,name, ocmaxmem) !! data  conj grad

       name="RS_s";      s_ = oc_clone(F,  x_,name, ocmaxmem)
       name="RS_Sm";    sm_ = oc_allocate(F, name, file%esize, nreg) !! model previous step
       name="RS_Sd";    sd_ = oc_clone(F, yy_,name, ocmaxmem) !! data  previous step

       name="RS_Td";     td_ = oc_clone(F, yy_,name, ocmaxmem)
       name="RS_Tm";     tm_ = oc_clone(F,  x_,name, ocmaxmem)

       if(ocmmovie) then
          name="RS_Mmovie"; mm_ = oc_clone(F, x_,name, ocmaxmem)
          call oc_adddim(mm_, ocniter)
       end if
       if(ocdmovie) then
          name="RS_Dmovie"; md_ = oc_clone(F, yy_,name, ocmaxmem)
          call oc_adddim(md_, ocniter)
       end if

       if( present(res_)) then
          call oc_shapeheader(res_,4,(/ocniter/),(/0./),(/1./))
       end if

       call oc_assign( rm_, real(rZERO), ocmaxmem)                                !! Rm =   0
       call oc_assign( sm_, real(rZERO), ocmaxmem)                                !! Sm =   0
       call oc_assign( gm_, real(rZERO), ocmaxmem)                                !! Gm =   0

       call oc_linear( rd_, (/yy_/), (/-rONE/), ocmaxmem)                         !! Rd = - d
       if(present(W)) then
         name="rt_";     rt_ = oc_clone(F, rd_,name, ocmaxmem)         !! data  residual
          if (ocimpi==0) stat=oc_scale(F,F,rd_,rt_,op1,op2,op3,op4,op5,op6,op7,op8,op9)
				 stat = W( F,F, rt_,rd_,op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Rd = -Wd
			 end if	
    end if

    if(ocimpi==0) call oc_scale_init(oceps,ocmaxmem)

    if(present(x0_)) then                                               !! Rd = WL x0 - Wd
       if(present(W)) then                                              !! Rm = eA x0
          if(.true.)    stat =        L( F, F, x0_,td_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!        L x0
          if(ocimpi==0) stat =        W( F, F, td_,rd_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Rd = W L x0
          if(ocimpi==0) stat =        A( F, F, x0_,tm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!        A x0
          if(ocimpi==0) stat = oc_scale( F, F, tm_,rm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Rm = e A x0
       else
          if(.true.)    stat =        L( F, F, x0_,rd_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Rd =   L x0
          if(ocimpi==0) stat =        A( F, F, x0_,tm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!        A x0
          if(ocimpi==0) stat = oc_scale( F, F, tm_,rm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Rm = e A x0
       end if
       if(ocimpi==0) call oc_linear( x_, (/x0_/),(/rONE/), ocmaxmem)                    !! x  = x0
    else
       if(ocimpi==0) call oc_assign( x_ , real(rZERO), ocmaxmem)                        !! x  =  0
    end if

    forget=T; !if(present(x0_)) forget=F
    do i = 1, ocniter
       if(ocimpi==0 .and. ocverb) call message('iteration',i)
       !----------------------------------------------------------------       
       if(ocimpi==0 .and. ocverb) call message('g= L*W* Rd + eA* Rm')
       if(present(W)) then
          if(ocimpi==0) stat =        W( T, F, td_,rd_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!        W* Rd
          if(.true.)    stat =        L( T, F,  g_,td_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! g  = L*W* Rd
          if(ocimpi==0) stat =        A( T, F, tm_,rm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!        A* Rm
          if(ocimpi==0) stat = oc_scale( T, T,  g_,tm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! g += e A* Rm
       else
          if(.true.)    stat =        L( T, F,  g_,rd_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! g  = L*   Rd
          if(ocimpi==0) stat =        A( T, F, tm_,rm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!        A* Rm
          if(ocimpi==0) stat = oc_scale( T, T,  g_,tm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! g += e A* Rm
       end if
       !----------------------------------------------------------------
       if(present(k_)) then
          if(ocimpi==0) call oc_mask(k_,g_,ocmaxmem) !! don't change known
       end if
       !----------------------------------------------------------------       
       if(ocimpi==0 .and. ocverb) call message('Gd= W L g; Gm= eA g')
       if(present(W)) then
          if(.true.)    stat =        L( F, F,  g_,td_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!        L  g
          if(ocimpi==0) stat =        W( F, F, td_,gd_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Gd = W L  g
          if(ocimpi==0) stat = oc_scale( F, F,  g_,tm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!      e    g
          if(ocimpi==0) stat =        A( F, F, tm_,gm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Gm = e A  g
       else
          if(.true.)    stat =        L( F, F,  g_,gd_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Gd =   L  g
          if(ocimpi==0) stat = oc_scale( F, F,  g_,tm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !!      e    g
          if(ocimpi==0) stat =        A( F, F, tm_,gm_, op1,op2,op3,op4,op5,op6,op7,op8,op9) !! Gm = e A  g
       end if
       !----------------------------------------------------------------
       if(ocimpi==0 .and. ocverb) call message('descend')
       if(ocimpi==0) stat = S(forget,x_,g_,(/rd_,rm_/),(/gd_,gm_/),s_,(/sd_,sm_/),ocmaxmem)
       !----------------------------------------------------------------
       if(ocimpi==0) then
          if(ocmmovie) then
             if(ocverb) call message('mmovie append')
             call oc_append(mm_,  x_)
          end if
          if(ocdmovie) then
             if(ocverb) call message('dmovie append')
             call oc_append(md_, rd_)
          end if
          !----------------------------------------------------------------
          if(present(res_)) then
             res = oc_norm((/rd_/), ocmaxmem)
             if( present(res_)) call sep_write(res, res_)
          end if
          !----------------------------------------------------------------
          if(stat ==1) then
             if( present(res_)) call oc_shapeheader(res_,4,(/i/),(/0./),(/1./))
             call message('DESCENT STOPPED')
             exit !! can't descend anymore
          end if
          !----------------------------------------------------------------
       end if
       forget = F !! forget=false except for i=1
    end do !! end iterations

  end subroutine oc_mpi_solverreg
  !----------------------------------------------------------------
end module oc_mpi_solverreg_mod


