module oc_off_pad
  use sep
  use oc_file_mod
  use oc_adjnull_mod
  use mpi_sep
  implicit none

  integer, private :: true_offx,true_offy,beg_x,beg_y,rampendx,rampendy
	real, private    :: halfpi
  integer,            private :: maxmem

contains

  !----------------------------------------------------------------
  subroutine oc_offpad_init( trueoffx,trueoffy,begx,begy,maxmem_in)
  	integer, intent(in) :: trueoffx,trueoffy,begx,begy
    integer,            optional   :: maxmem_in
    maxmem=0; if(present(maxmem_in)) maxmem=maxmem_in
    true_offx=trueoffx
    true_offy=trueoffy
		beg_x=begx
		beg_y=begy
		rampendx=true_offx/4
		if(true_offy .gt. 8) then
		  rampendy=true_offy/4
		else
			rampendy=1
		end if
		halfpi=1.5707963
  end subroutine oc_offpad_init
  !----------------------------------------------------------------
  integer function oc_offpad(adj,add, x_,yy_, op1,op2,op3,op4,op5,op6,op7,op8,op9) result(stat)
    logical,               intent(in)  :: adj,add
    character(len=128),    intent(in)  :: x_,yy_
    integer, external,     optional    :: op1,op2,op3,op4,op5,op6,op7,op8,op9
    type(fileinfo)                     :: filex,filey
    integer                            :: i,j,ithread,ierr
    integer, external :: sreed
		complex, dimension(:,:,:,:,:), allocatable :: resid_in, resid_out 


	  call MPI_COMM_RANK(MPI_COMM_WORLD, ithread,ierr)
if(ithread==0) then

write(0,*)"off_pad"

    call oc_adjnull(adj,add,x_,yy_)

    call oc_allocatefile(filex, x_,  maxmem); call sseek(x_, 0,0)
    call oc_allocatefile(filey, yy_, maxmem); call sseek(yy_,0,0)
    call oc_checksimilarity(filex,filey, "oc_weight")

		allocate(resid_in(filey%n(1),filey%n(2),filey%n(3),filey%n(4),filey%n(5)))
		allocate(resid_out(filex%n(1),filex%n(2),filex%n(3),filex%n(4),filex%n(5)))

		resid_out=0
		resid_in=0

write(0,*)"checking off_pad",beg_x,beg_y,true_offx,true_offy,size(resid_in),size(resid_out),rampendx,rampendy

    if(.not. adj) then  !! x=w y

             write(0,*)"off_pad: forward",trim(x_)
             ierr=sreed(x_, resid_in, 8*filex%n(1)*filex%n(2)*filex%n(3)*filex%n(4)*filex%n(5))

write(0,*)"in forward1",sum(cabs(resid_in)),sum(cabs(resid_out)),ierr
write(0,*)"in forward1a",sum(cabs(resid_in(1:15))),sum(cabs(resid_out(1:15)))

						 resid_out(:,:,beg_x:true_offx,beg_y:true_offy,:)=resid_in(:,:,beg_x:true_offx,beg_y:true_offy,:)

						 do i=1,rampendx
								resid_out(:,:,beg_x+(i-1),:,:)=resid_out(:,:,beg_x+(i-1),:,:)*((cos(-halfpi+(halfpi/rampendx)*(i-1)))**2)	
						 end do

!						 do j=1,rampendy
!								resid_out(:,:,:,beg_y+(j-1),:)=resid_out(:,:,:,beg_y+(j-1),:)*((cos(-halfpi+(halfpi/rampendy)*(j-1)))**2)	
!						 end do

write(0,*)"in forward2",sum(cabs(resid_in)),sum(cabs(resid_out))
write(0,*)"in forward2a",sum(cabs(resid_in(1:15))),sum(cabs(resid_out(1:15)))

						 call sep_write(resid_out, yy_)	
    else                     !! y=w x  

             write(0,*)"off_pad: adjoint; weighting residual so offset padding is zero"
             ierr= sreed(yy_, resid_out, 8*filex%n(1)*filex%n(2)*filex%n(3)*filex%n(4)*filex%n(5))

write(0,*)"FUC",sum(cabs(resid_out)),ierr
						 resid_in(:,:,beg_x:true_offx,beg_y:true_offy,:)=resid_out(:,:,beg_x:true_offx,beg_y:true_offy,:)

						 do i=1,rampendx
								resid_in(:,:,beg_x+(i-1),:,:)=resid_in(:,:,beg_x+(i-1),:,:)*((cos(-halfpi+(halfpi/rampendx)*(i-1)))**2)	
						 end do

!						 do j=1,rampendy
!								resid_in(:,:,:,beg_y+(j-1),:)=resid_in(:,:,:,beg_y+(j-1),:)*((cos(-halfpi+(halfpi/rampendy)*(j-1)))**2)	
!						 end do

write(0,*)"in adjoint",sum(cabs(resid_in)),sum(cabs(resid_out))
write(0,*)"in adjoint1",sum(cabs(resid_in(1:15))),sum(cabs(resid_out(1:15)))

             call sep_write(resid_in, x_)
    end if

		deallocate(resid_in,resid_out)
    call oc_deallocatefile(filex); call sseek(x_, 0,0)
    call oc_deallocatefile(filey); call sseek(yy_,0,0)
end if
    stat = 0
  end function oc_offpad

end module oc_off_pad
