! Copyright 1999 Stanford University
!All rights reserved
!
!<
!build_filter
!
!Usage:
!call build_filter (arg1, arg2 arg3)
!
!Input Parameters:
!
!
!Output Parameters:
!
!
!
!Description:
!
!
!
!>
!-------------------------------------------------
!
!Author: Robert Clapp, ESMB 463, 7230253
!
!Date Created:Tue Jun 22 14:11:11 PDT 1999
!
!Purpose:
!
!
module build_filter_mod 
  use radial_mod
  use createnhelixmod
  use steering
  use lapfac
  implicit none
  integer,private,save  :: n(2)  !data size
  integer,private,save  :: j(2)  
  !sampling of filters in cartesian space
  integer,private,save  :: center(2),na(2)  
  !box size and location of zero lag
  real,private,save     :: o(2),d(2)  !sampling of data
  real,private,save     :: hwidth,amp 
  !width and amplitude of radial smoother
  real,private,save     :: dtheta,dtau !sampling for radial filters
  logical,private,save  :: radial,steer  !whether or no
  contains
!subroutine to initialize or change any of the building parameters
  subroutine init_build_filters(n_in,j_in,center_in,na_in,o_in,d_in&
    &,hwidth_in,amp_in,radial_in,steer_in,dtau_in,dtheta_in)
    integer, optional :: n_in(2),j_in(2),center_in(2),na_in(2)
    real,optional     :: o_in(2),d_in(2),hwidth_in,amp_in,dtau_in&
      &,dtheta_in
    logical,optional  :: radial_in,steer_in
    if (present(n_in)) then
      n=n_in
    end if
    if (present(j_in)) then
      j=j_in
    end if
    if (present(center_in)) then
      center=center_in
    end if
    if (present(na_in)) then
      na=na_in
    end if
    if (present(o_in)) then
      o=o_in
    end if
    if (present(d_in)) then
      d=d_in
    end if
    if (present(hwidth_in)) then
      hwidth=hwidth_in
    end if
    if (present(amp_in)) then
      amp=amp_in
    end if
    if (present(radial_in)) then
      radial=radial_in
    end if
    if (present(steer_in)) then
      steer=steer_in
    end if
    if (present(dtau_in)) then
      dtau=dtau_in
    end if
    if (present(dtheta_in)) then
      dtheta=dtheta_in
    end if
  end subroutine 
  subroutine build_2d_filter(filt,rough,ncoef)
    type(nfilter)        :: filt,rough  !filter and rough
    integer              :: ncoef       !number of coeficient
!local variables
    integer              :: n123
    integer,allocatable  :: map(:),nh(:)
    real,allocatable     :: p(:)
    integer              :: i2,i1,iout(2),iref,npts(2),resamp,i
    real                 :: pos(2),theta,tau,reg
    integer              :: itau,itheta
    integer,allocatable  :: found(:,:)
    integer              :: ntheta,ntau,noff
    logical              :: good
!FIRST BUILD FILTER
    if (.not. radial) then
      call build_cartesian_space(filt,ntau,noff,ncoef)
    else
      call build_radial_space(filt,ntau,ntheta,ncoef)
    end if 
   !write(0,*) "ntau noff ",ntau,ntheta,noff,radial
!NEXT BUILD ROUGHNER
    if (radial) then
      call create_radial_smooth(ncoef,ntau,ntheta,rough)
    else if (steer) then
      call create_steer_rough(ncoef,ntau,noff,rough)
    else
      call create_lap_rough(ncoef,ntau,noff,rough)
    end if
  end subroutine 
  subroutine build_cartesian_space(filt,ntau,noff,ncoef)
    type(nfilter)         :: filt
    integer               :: ntau,noff,ncoef
    integer               :: i1,i2,i,iref,iout(2)
    integer,allocatable   :: map(:)
    allocate(map(product(n)))
    ntau=ceiling(real(n(1))/real(j(1)))  !the number of time filters
    noff=ceiling(real(n(2))/real(j(2)))  
    !the number of offset  filters
    !write(0,*) "j is",j
    do i2=1,n(2)
      iout(2)=ceiling(real(i2)/real(j(2)))
      do i1=1,n(1)
        iout(1)=ceiling(real(i1)/real(j(1)))
        i=i1+n(1)*(i2-1)
        iref=iout(1)+ntau*(iout(2)-1)
        map(i)=iref
      end do
    end do 
		!write(0,*) "ss1",na,n,size(map)
    filt=createnhelix(n(1:2),center,(/0,0/),na,map)  
		!write(0,*) "ss2"
    !create prediction filter
    ncoef=size(filt%hlx(1)%lag)
    deallocate(map)     !record number of coef in f
		!write(0,*) "made it ithis far"
  end subroutine 
  subroutine build_radial_space(filt,ntau,ntheta,ncoef)
    integer               :: ntau,ntheta,ncoef
    type(nfilter)         :: filt
    real,allocatable      :: p(:)
    integer,allocatable   :: map(:)
    integer,allocatable   :: itau(:),itheta(:)
    real                  :: otheta,dspace,max1,max2,dist,pos2,five_ang&
      &,v
    integer               :: i1,i2,i
    allocate(p(product(n)),map(product(n)))
    allocate(itau(product(n)),itheta(product(n)))
    call find_p(n,o,d,(/1,1/),p,.true.,90.) !we just found the angles
    if (d(2)<1.) then
      v=8.
    else
      v=8000.
    end if 
    max1=1./d(1)+1.5	
    max2=abs(v/d(2)+1.5)
    five_ang=atan(max2/max1)*45/atan(1.)
    where(p< -five_ang)
      p=-five_ang
end where
    where(p> five_ang)
      p=five_ang
end where
    otheta=minval(p)	
    ntheta=ceiling((maxval(p)-otheta)/dtheta)+1
!THE OTHER DIMENSION IS ACTUALLY A DISTANCE FROM THE ORIGIN THING
    dspace=dtau/d(1)
    max1=n(1)
    max2=max(abs(o(2)),abs(o(2)+d(2)*(n(2)-1)))/d(2)
    ntau=ceiling(sqrt(max1**2+max2**2)/dspace)+1
    !write(0,*) "ntheta ntau", ntheta,ntau,dspace
!loop over
    i=0
    do i2=1,n(2)
      pos2=o(2)/d(2) + i2-1
      do i1=1,n(1)
        i=			i+1
        itheta(i)=nint(real(p(i)-otheta)/dtheta)+1
        dist=sqrt(real(i1**2 + pos2**2))/dspace
        itau(i)=nint(dist)+1
      end do
    end do 
    call srite("theta.H",real(itheta),size(itheta)*4)
    call srite("tay.H",real(itau),size(itau)*4)
    map=itau+(itheta-1)*ntau
    deallocate(itau,itheta)
    filt=createnhelix(n(1:2),center,(/0,0/),na,map)  
    !create prediction filter
    ncoef=size(filt%hlx(1)%lag)
    deallocate(map)     !record number of coef in f
  end subroutine 
  subroutine create_radial_smooth(ncoef,ntau,ntheta,rough)
    type(nfilter)         :: rough
    integer               :: ntau,ntheta,ncoef,npts(2)
    integer               :: i1,i2
    integer,allocatable   ::  nh(:),map(:)
    integer               ::  npts(2),i2,i,resamp,i1
    npts=(/ntau,ntheta/)
    allocate(nh(product(npts)))
    allocate(map(ncoef*product(npts)))
    nh=min(ceiling(hwidth*2.),product(npts))
    do i2=1,product(npts)
!apply the same filter to every filter coeficient
      do i1=1,ncoef 
        i=i1+(i2-1)*ncoef
        map(i)=i2
      end do
    end do 
    call nallocate(rough,nh,map)
    if (nh(1).ne.1) then
!if our filter is small apply just the idenity for reg
      call steer_init(hwidth,amp,npts(1))
      do i=1,product(npts)
        call steering2(0.,rough%hlx(i))
      end do
    else
      rough%hlx(1)%lag=1
      rough%hlx(1)%flt=0.
    end if 
    resamp=ncoef
    do i=1,size(rough%hlx)
      rough%hlx(i)%lag=rough%hlx(i)%lag*resamp
    end do 
    deallocate(nh,map)
  end subroutine 
  subroutine create_steer_rough(ncoef,ntau,noff,rough)
    integer               ::  ntau,noff,ncoef
    type(nfilter)         ::  rough
    integer,allocatable   ::  nh(:),map(:)
    real,allocatable      ::  p(:)
    integer               ::  npts(2),i2,i,resamp,i1
    npts=(/ntau,noff/)
    allocate(nh(product(npts)))
    allocate(map(ncoef*product(npts)))
    nh=min(ceiling(hwidth*2.),product(npts))
    allocate(p(product(npts)))   !THE SLOPES
    do i2=1,product(npts)
!apply the same filter to every filter coeficient
      do i1=1,ncoef 
        i=i1+(i2-1)*ncoef
        map(i)=i2
      end do
    end do 
    call nallocate(rough,nh,map)
    if (nh(1).ne.1) then
!if our filter is small apply just the idenity for reg
      call find_p(npts,o,d,j,p,.false.,4.)
      call steer_init(hwidth,amp,npts(1))
      do i=1,size(p)
        call steering2(p(i),rough%hlx(i))
      end do 
      deallocate(p)
    else
      rough%hlx(1)%lag=1
      rough%hlx(1)%flt=0.
    end if 
    resamp=ncoef
    do i=1,size(rough%hlx)
      rough%hlx(i)%lag=rough%hlx(i)%lag*resamp
    end do 
    deallocate(nh,map)
  end subroutine 
  subroutine create_lap_rough(ncoef,ntau,noff,rough)
    integer               ::  ntau,noff,ncoef
    type(nfilter)         ::  rough
    integer,allocatable   ::  nh(:),map(:)
    integer               ::  npts(2),resamp,i
    npts=(/ntau,noff/)
    allocate(nh(product(npts)))
    nh=ncoef
    allocate(map(ncoef*product(npts)))
    map=1
    nh=min(6,npts(1))*2  !coeficient size
    call nallocate(rough,nh,map)
    rough%hlx(1)=lapfac2(.01,npts(1),nh(1)/2)
    resamp=1
    deallocate(nh,map)
    do i=1,size(rough%hlx)
      rough%hlx(i)%lag=rough%hlx(i)%lag*resamp
    end do
  end subroutine 
!if(.not. radial){
!	npts(1:2)=ceiling(real(n(1:2))/real(j(1:2)))  #the number of filters
!
!  do i2=1,n(2){
!    iout(2)=ceiling(real(i2)/real(j(2)))
!    do i1=1,n(1){
!      iout(1)=ceiling(real(i1)/real(j(1)))
!      i=i1+n(1)*(i2-1)
!      iref=iout(1)+npts(1)*(iout(2)-1)
!      map(i)=iref
!    }
!  }
!}
!else{ #create radial map
!	}
!	i=0
!	do i2=1,ntheta{
!		do i1=1,ntau{
!			if(found(itau,itheta)==-1){ #we have a match
!				i+=1
!				found(itau,itheta)=i
!			}
!		}
!	}
!	i=0
!	do i2=1,n(2){
!		pos(2)=o(2)+d(2)*(i2-1)
!		do i1=1,n(1){
!			i+=1
!			pos(1)=o(1)+d(1)*(i1-1)
!			tau=pos(1)
!			theta=atan(real(i2)/real(i1))
!			itau=floor(tau/dtau+1.5)
!			itheta=floor(theta/dtheta+1.5)
!			map(i)=found(itau,itheta)
!		}
!	}
!}
!filt=createnhelix(n(1:2),center,(/0,0/),na,map)  #create prediction filter
!ncoef=size(filt%hlx(1)%lag); deallocate(map)     #record number of coef in f
!#########################
!      ROUGHNER          #
!#########################
!if(.not.radial){
!  allocate(nh(product(npts)));nh=ncoef;
!  allocate(map(ncoef*product(npts)))
!  if(steer){ #DOING RADIAL SMOOTHING
!		nh=min(ceiling(hwidth*2.),product(npts))
!
!    allocate(p(product(npts)))   #THE SLOPES
!
!
!    do i2=1,product(npts){ #apply the same filter to every filter coeficient
!      do i1=1,ncoef{
!        i=i1+(i2-1)*ncoef
!        map(i)=i2
!      }
!    }
!		call nallocate(rough,nh,map)
!		if(nh(1)!=1){  #if our filter is small apply just the idenity for reg
!    	call find_p(npts,o,d,j,p,.true.,90.)
! 	   call steer_init(hwidth,amp,npts(1))
! 	   do i=1,size(p)
! 	     call steering2(p(i),rough%hlx(i))
! 	   deallocate(p)
!		}
!		else{
!			rough%hlx(1)%lag=1
!			rough%hlx(1)%flt=0.
!		}
!		resamp=ncoef
!  }
!  else{  #In this case we are using a laplacian regularizer
!    map=1
!		nh=min(6,npts(1))*2  #coeficient size
!    call nallocate(rough,nh,map)
!    rough%hlx(1)=lapfac2(.01,npts(1),nh(1)/2)
!		resamp=1
!  }
!}
!else{
!  allocate(map(ncoef*maxval(found)))
!	i1=0;i2=0
!	do itheta=1,ntheta{
!		do itau=1,ntau{
!			if(found(itau,itheta)!=0){
!				i2+=1
!				do i=1,ncoef{
!					i1+=1
!					map(i1)=i2
!				}
!			}
!		}
!	}
!	allocate(nh(i2));nh=ceiling(hwidth*2.)
!	call nallocate(rough,nh,map)
!	reg=0.
!	do i2=1,nh(found(itau,itheta)){
!		i1=itheta-ceiling(hwidth)-1+i2
!		reg+=abs(hwidth-abs(real(i1-itheta)))/hwidth
!	}
!	reg=-amp/reg
!
!
!	do itheta=1,ntheta{
!		do itau=1,ntau{
!			if(found(itau,itheta)!=0){
!				good=.true.
!				do i2=1,nh(found(itau,itheta)){
!					i1=itheta-ceiling(hwidth)-1+i2
!					if(itheta==ntheta  || i1<1 || i1>ntau) good=.false.
!					else if(found(i1,itheta+1) ==0) good=.false.
!					else{
!						rough%hlx(found(itau,itheta))%lag(i2)=found(i1,itheta+1)
!						rough%hlx(found(itau,itheta))%flt(i2)=reg*
!							abs(hwidth-abs(real(i1-itheta)))/hwidth
!					}
!				}
!				if(.not. good){
!					 rough%hlx(found(itau,itheta))%lag(i2)=1.
!           rough%hlx(found(itau,itheta))%flt(i2)=0.
!				}
!			}
!		}
!	}
!	deallocate(found)
!	resamp=ncoef
!}
!  deallocate(map,nh)
!
!
!#WE DON'T CALL NBOUND ON THE ROUGHNER (maybe we should if it is big)
!#definite problem error
!# 	call nboundn(npts,rough)
!
!	#rescale lag coeficient by the number of coeficients in the
!  #prediction filter
!  do i=1,size(rough%hlx)
!    rough%hlx(i)%lag=rough%hlx(i)%lag*resamp
!}
end module 
