module wei_obj_D
  use wei_util
  use wei_down
  use wei_run
  use wei_external
  use wei_type
  !---------------
  implicit none
contains
  !--------------------------------------------------------------
  !
  !  Function to initialize a data object either from
  ! 
  !    a file or from then general parameters
  !
  !
  !
  logical function wei_data_init(X,tag)
    type(wedata),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    wei_data_init=F

    X%tag = tag
    if(wei_dfile) then
       if(.not. data_history_init(X)) then
         write(0,*) "trouble initializing data_history"
         return
       end if
    else
       if(.not. data_param_init(X)) then
         write(0,*) "trouble initializing data_paramater"
         return
       end if
    end if

    wei_data_init=T

  end function wei_data_init

  !----------------------------------------------------------------
  !
  !
  !  Create an auxilary data file
  !
  logical function wei_aux_data_init(X,tag)
    type(wedata),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    wei_aux_data_init=F

    X%tag = tag
    if(.not. data_scratch_init(X)) then
      write(0,*) "trouble initializing scratch data file"
      return
    end if

    wei_aux_data_init=T

  end function wei_aux_data_init
  !----------------------------------------------------------------
  !
  !  Initialize a tag from history
  !   check to make sure that is dimensions don't disagree
  !    with the general params
  ! 
  logical function data_history_init(X)
    type(wedata), intent(inout) :: X

#ifdef DEBUG
    character(len=128), parameter :: unit='data_history_init'
    if(run%debug) call in(unit)
#endif

    call auxinout(X%tag)
    data_history_init=F
    
    
    X%from = "history"

    if(.not. data_setup(X)) then
      write(0,*) "trouble setting up data"
      return
    end if

    data_history_init=T

#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif

    
  end function data_history_init

  !----------------------------------------------------------------
  !
  !
  !
  logical function create_data_file(X,tag)
    type(wedata), intent(in) :: X
    character(len=*)         :: tag
    real :: a 
    integer :: i
       create_data_file=F 
       call auxscr(tag)
       call sep3d_write_description(tag,down%dsep)
       i= srite(tag,a,4)
       call auxclose(tag) 
       call auxinout(tag)
       create_data_file=T 
   end function 


  !----------------------------------------------------------------
  !
  !
  !
  logical function data_param_init(X)
    type(wedata), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='data_param_init' 
    if(run%debug) call in(unit)
#endif

    data_param_init=F
    X%from = "parameter"
     if(run%m_ith==0) then
      if(.not. create_data_file(X,X%tag)) then
        write(0,*) "trouble creating data file"
        return
      end if
    end if
    if(.not. data_setup(X)) then
      write(0,*) "trouble setting up data"
      return
    end if

    data_param_init=T
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function


  subroutine data_assign(X,val)
   type(wedata) :: X
   complex :: val
   logical ::a

    a=file_val_assign(X%tag,val)
  end subroutine
  !----------------------------------------------------------------
  !
  ! data_scratch_init
  !
  !  Create a data scratch file
  !
  !
  logical function data_scratch_init(X)
    type(wedata), intent(inout) :: X
    integer :: n
#ifdef DEBUG
    character(len=128), parameter :: unit='data_scratch_init' 
    if(run%debug) call in(unit)
#endif

#ifdef SEP_MPI
    X%tag_0=X%tag
    X%aw_local%n=size(down%iw_first)*down%nws
    if(.not. make_local_tag(X%tag,x%tag_0,run%tag_num)) then
       write(0,*) "trouble making tag ",trim(X%tag)
       return
    end if
#endif

    data_scratch_init=F
    X%from = "parameter";
    n=down%dsep%n(5)
    down%dsep%n(5)=size(down%iw_first)*down%nws
    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble  creating data_file"
      return
    end if
    down%dsep%n(5)=n
    if(.not. data_setup(X,.false.)) then
      write(0,*) "trouble setting up data"
      return
    end if
    data_scratch_init=T

#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_scratch_init
  !----------------------------------------------------------------
  ! 
  ! 
  ! 
  !  The core  routine to setup data
  ! 
  ! 
  logical function data_setup(X,allocit)
    type(wedata), intent(inout) :: X
    logical :: zero,alloc
    logical, optional :: allocit
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='data_setup' 
    if(run%debug) call in(unit)
#endif
    data_setup=F

   alloc=T;if(present(allocit)) alloc=allocit

     zero=.true.;if(X%from(1:3)=="his") zero=.false.

    X%bites = 8*product(down%wsep%n(1:4))*down%wsep%n(6)



    !! FREQUENCY blocking
    if( (1.0*X%aw_local%n)/down%nws /= X%aw_local%n/down%nws ) then
       call data_report(X)
       call seperr('D.aw.n is not a multiple of nws')
    end if
    down%nwb = down%nw_local !/ down%nws




    !allocate array
    if(run%mem_save < 2 .and.  alloc) then
      allocate(X%bin(down%wsep%n(1),down%wsep%n(2),down%wsep%n(3),down%wsep%n(4),down%nws),stat=er)
      call add_mem(size(X%bin))
      if(er/=0) call seperr('cant allocate D.bin')
      X%bin=(0.0,0.0)
    else  if(run%mem_save > 1) then
      X%bin=>down%block
      X%bin=(0.0,0.0)
    end if



    if(zero .and. run%m_nth==1) then
      if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
        write(0,*) "trouble setting ",trim(X%tag)
        return
      end if
    end if
   if(.not.run%mpi) then
     X%aw_local%n=down%dsep%n(5)
     X%aw_local%o=down%dsep%o(5)
     X%aw_local%d=down%dsep%d(5)
   end if


   data_setup=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_setup

  !----------------------------------------------------------------
  !
  !   Setup data structure for local files if
  !     we are running MPI
  !
  !

  logical function mpi_data_setup(X)
    type(wedata), intent(inout) :: X
    integer ::n,st
    logical :: zero
    mpi_data_setup=F

     zero=.true.;
!    if(X%from(1:3)=="his") zero=.false.
    if(down%adj) zero=.false.



    X%tag_0=X%tag
    X%aw_local%n=size(down%iw_first)*down%nws
    if(.not. make_local_tag(X%tag,x%tag_0,run%tag_num)) then
       write(0,*) "trouble making tag ",trim(X%tag)
       return
    end if

    n=down%dsep%n(5)
    down%dsep%n(5)=size(down%iw_first)*down%nws
    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble setting up mpi data"
      return
    end if
    down%dsep%n(5)=n


    if(.not. zero) then 
         call distribute_data(X)
    else
       if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
          write(0,*) "trouble setting ",trim(X%tag_0)
          return
       end if
    end if

    mpi_data_setup=T
  end function

  !----------------------------------------------------------------
  !
  ! Information on the data
  !
  !
  subroutine data_report(X)
    type(wedata), intent(in) :: X
    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    Mx: ",down%dsep%n(1), down%dsep%o(1), down%dsep%d(1)
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    My: ",down%dsep%n(2), down%dsep%o(2), down%dsep%d(2)
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    Hx: ",down%dsep%n(3), down%dsep%o(3), down%dsep%d(3)
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    Hy: ",down%dsep%n(4), down%dsep%o(4), down%dsep%d(4)
    write(0,'(a,i3,a,i3,a,f10.2,f10.3,a,i3,a)') "         *     W: ",down%dsep%n(5),&
    "(",X%aw_local%n,")",  down%dsep%o(1),  down%dsep%d(2)," ->",down%nwb," blocks"
    if(associated(X%bin)) then
       write(0,'(a,f9.2,a,f9.2,a)        ') "         * block: ",X%bites/1024.," / ",8*size(X%bin)/1024.," Kbytes"
    end if
    write(0,*)'        ************************************************'
  end subroutine data_report
  !--------------------------------------------------------------
  ! 
  ! Read the data 
  ! 
  !
  logical function data_read(X,k)
    type(wedata), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='data_read'
    if(run%debug) call in(unit)
#endif
#ifdef SEP_TIMER
  if(run%m_ith==0)  call start_timer_num(timer%io_t)
#endif

    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%bites/= sreed(X%tag, X%bin, X%bites)) then
      write(0,*) "trouble reading block",k," thread=",run%m_ith
      write(0,*) "seeking bytes",X%bites," block",k-1
      return
    end if

    data_read=T

#ifdef SEP_TIMER
  if(run%m_ith==0)  call stop_timer_num(timer%io_t)
#endif
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_read
  !----------------------------------------------------------------
  logical function data_write(X,k)
    type(wedata), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='data_write' 
    if(run%debug) call in(unit)
#endif
#ifdef SEP_TIMER
  if(run%m_ith==0)  call start_timer_num(timer%io_t)
#endif
    data_write=F 

    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%bites/= srite(X%tag, X%bin, X%bites)) then
      write(0,*) "trouble writing data block",k
      return
    end if

     call to_history("ifreq",k,X%tag)

    data_write=T 
#ifdef SEP_TIMER
  if(run%m_ith==0)  call stop_timer_num(timer%io_t)
#endif
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_write
  !----------------------------------------------------------------
  !
  ! 
  !
  !
  subroutine distribute_data(X)
    type(wedata), intent(inout) :: X
    integer, external :: sseek
    integer :: ierr

#ifdef SEP_MPI
    if(run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(x%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)
       if(0/=mpi_sep_tag_distribute(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
       down%nwp, down%iw_thread)) &
       call seperr("trouble distributing data ");
    end if
#endif
  end subroutine distribute_data
  !----------------------------------------------------------------
  !
  ! 
  !
  !
  subroutine combine_data(X)
    type(wedata), intent(inout) :: X
    integer :: ierr,i
    integer, external :: sseek
#ifdef SEP_MPI
     i=0;if(down%add) i=1
    if( .not. down%adj .and. run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(x%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)

       if(0/=mpi_sep_tag_combine(run%m_ith,X%tag,0,X%tag_0, &
       X%bites/4,down%nwp,down%iw_thread,i)) &
       call seperr("trouble combining data ");
    end if
     if(run%unlink) call aux_unlink(trim(X%tag))
     call auxclose(trim(X%tag))
#endif
  end subroutine combine_data

  !----------------------------------------------------------------
  !
  !  Copy data
  !
  !
  subroutine data_copy(X,Y)
    type(wedata), intent(inout) :: X
    type(wedata), intent(inout) :: Y
    integer :: iwb,nblock,ndone,n123
    logical :: res
#ifdef DEBUG
    character(len=128), parameter :: unit='data_copy'
    if(run%debug) call in(unit)
#endif


    n123=size(X%bin)
    iwb= sseek(trim(X%tag),0,0)
    iwb= sseek(trim(y%tag),0,0)
    do iwb=1,down%nwb
       ndone=0
       do while(ndone < n123)
         nblock=min(n123-ndone,size(small_buf,1))
        if(8*nblock.ne. sreed(trim(X%tag),small_buf,8*nblock)) then
            write(0,*) iwb,"=iwb done=",ndone,trim(X%tag),"=tag nblock=",nblock
            write(0,*) ndone,"=ndone n123=",n123
            call seperr("trouble reading from tag (copy)")
         end if
         if(8*nblock.ne. srite(trim(y%tag),small_buf,8*nblock)) then
            write(0,*) ndone,"=ndone n123=",n123
            write(0,*) iwb,"=iwb done=",ndone,trim(y%tag),"=tag nblock=",nblock
            call seperr("trouble writing to tag (copy)")
         end if
         ndone=ndone+nblock
       end do
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_copy


  !----------------------------------------------------------------
  ! 
  !  Add to data structures
  ! 
  ! 
  subroutine data_add(X,Y)
    type(wedata), intent(inout) :: X
    type(wedata), intent(inout) :: Y
    integer :: iwb,n123,ndone,nblock,i
    logical :: res
#ifdef DEBUG
    character(len=128), parameter :: unit='data_add'
    if(run%debug) call in(unit)
#endif
    n123=size(X%bin)
    iwb= sseek(trim(X%tag),0,0)
    iwb= sseek(trim(y%tag),0,0)
    do iwb=1,down%nwb
       ndone=0
       do while(ndone < n123)
         nblock=min(n123-ndone,size(small_buf,1))
         if(8*nblock.ne. sreed(trim(X%tag),small_buf,8*nblock)) then
            write(0,*) iwb,"=iwb done=",ndone,trim(X%tag),"=tag nblock=",nblock
             write(0,*) "read ",nblock," of ",n123
            call seperr("trouble reading from tag (add)")
         end if
         if(8*nblock.ne. sreed(trim(y%tag),small_buf(:,2),8*nblock)) then
            write(0,*) iwb,"=iwb done=",ndone,trim(X%tag),"=tag nblock=",nblock
             write(0,*) "read2",nblock," of ",n123
            call seperr("trouble reading from tag (add)")
         end if
         small_buf(:,1)=small_buf(:,1)+small_buf(:,2)
         i=sseek(trim(y%tag),-8*nblock,1)
         if(8*nblock.ne. srite(trim(y%tag),small_buf,8*nblock)) then
            write(0,*) iwb,"=iwb done=",ndone,trim(y%tag),"=tag nblock=",nblock
            call seperr("trouble writing to tag (add)")
         end if
         ndone=ndone+nblock
       end do
    end do

!    do iwb=1,down%nwb
!       res= data_read (X,iwb)
!       res= data_read (Y,iwb)
!       Y%bin = Y%bin + X%bin
!       res= data_write(Y,iwb)
!    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_add
  !----------------------------------------------------------------  
end module wei_obj_D
