module wei_obj_R
  use wei_util
  use wei_external
  use wei_run
  use wei_down
  use wei_type
  use wei_lagrange
  !---------------
  implicit none
  integer, private, save,pointer  :: image_master(:)
  integer, private, save  :: image_comm_group,ihx_beg,ihy_beg,ihx_end,ihy_end
  logical, private, save  :: image_lead_thread


contains
  !----------------------------------------------------------------
  !
  !
  ! Image initializing routine
  !
  !
  logical function  wei_image_init(X,tag)
    type(weimag),       intent(inout) :: X
    character(len=128), intent(in)    :: tag
    integer ::ierr,i
    real :: f

    wei_image_init=F

    X%tag = tag


#ifdef SEP_MPI
    call MPI_Barrier(MPI_COMM_WORLD,ierr)
#endif
    if(wei_rfile) then
       if(.not. image_history_init(X)) then
         write(0,*) "trouble initializing image_history"
         return
       end if       
    else
       if(.not. image_param_init(X)) then
         write(0,*) "trouble initializing image_parm"
         return
       end if       
    end if


    !CHECK TO SEE IF ONLY DOING ZERO OFFSET
    if(.not. compute_write_limits(ihx_beg,ihx_end,3)) then
      write(0,*) "trouble computing write limits for hx"
      return
    end if

    if(.not. compute_write_limits(ihy_beg,ihy_end,4)) then
      write(0,*) "trouble computing write limits for hy"
      return
    end if

    if(var%use_disk_buf) then
       if(.not. make_local_tag(X%tag_var,"R_var",run%tag_num)) then
          write(0,*) "trouble creating local tag" 
          return
       end if 
       call auxinout(X%tag_var)
       ierr= srite(X%tag_var,0.,4)
    end if

    wei_image_init=T
     
  end function wei_image_init
  

  logical function compute_write_limits(ib,ie,iaxis)
    integer :: ib,ie,iaxis
    real    :: f
    compute_write_limits=.false. 
if(.not. down%pcig) then
    if(abs(down%rsep%d(iaxis)-down%wsep%d(iaxis))/down%wsep%d(iaxis)>.001) then
      write(0,*) "d of image and wavefield must be the same"
      return
    end if
end if
      
    if(down%rsep%n(iaxis) /= down%wsep%n(iaxis)) then
      f=(down%rsep%o(iaxis)-down%wsep%o(iaxis))/down%wsep%d(iaxis)+1.
      ib=nint(f)
      if(abs(f-ib) > .001) then
        write(0,*) "begining of axis limit not integer of wavefield space"
        return
      end if
      f=(down%rsep%o(iaxis)+down%rsep%d(iaxis)*(down%rsep%n(iaxis)-1) &
          -down%wsep%o(iaxis))/down%wsep%d(iaxis)+1.
      ie=nint(f)
      if(abs(f-ie) > .001) then
        write(0,*) "end of axis limit not integer of wavefield space"
        return
      end if 
      if(ib < 1 .or. ie > down%wsep%n(iaxis)) then
        write(0,*) "illegal range ",ib,ie
        return 
      end if
    else
      ib=1;ie=down%wsep%n(iaxis)
    end if 
    compute_write_limits=.true. 
   end function

    
  !----------------------------------------------------------------
  !
  !
  ! Image initializing routine when the image file exists
  !
  !
  logical function image_history_init(X)
    type(weimag), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='image_history_init'
    if(run%debug) call in(unit)
#endif
    image_history_init=F

    call auxinout(X%tag)
    X%from = "history"

    X%esize=8;if(down%rsep%data_format(1:5)=="FLOAT") X%esize=4
    if(.not. image_setup(X)) then
      write(0,*) "trouble setting up image"
      return
    end if

    image_history_init=T

#ifdef DEBUG
    if(run%debug) call image_report(X)
    if(run%debug) call out(unit)
#endif
  end function image_history_init

  !----------------------------------------------------------------
  !
  ! Write out parameters for an image file
  !
  !
  logical function create_image_file(X,tag,main)
    type(weimag) :: X
    logical :: i_real,main
    integer :: i
    character(len=*) ::  tag
    create_image_file=F
      if(.not. main) call auxscr(tag)
      if(run%m_ith==0) call sep3d_write_description(tag,down%rsep)
        if(.not. main)  then
        call auxclose(tag)
        call auxinout(tag)
        end if

       i=srite(tag,0.,4)
       i=sseek(tag,0,0)
!       if(run%share_image) call auxclose(tag)
    create_image_file=T
  end function


  !----------------------------------------------------------------
  !
  ! Initialize an image file based on parameters
  !
  !
  logical function image_param_init(X)
    type(weimag), intent(inout) :: X
    logical :: image_real
#ifdef DEBUG
    character(len=128), parameter :: unit='image_param_init'
    if(run%debug) call in(unit)
#endif

    image_param_init=F
    X%from = "parameter"

    X%esize=8;if(down%rsep%data_format(1:5)=="FLOAT") X%esize=4


     
    if(run%m_ith==0) then 
      if(.not. create_image_file(X,X%tag,.true.)) then
        write(0,*) "trouble creating image"
        return
       end if
    end if
    if(.not. image_setup(X)) then
      write(0,*) "trouble  in image setup"
      return
    end if


     image_param_init=T

#ifdef DEBUG
    if(run%debug) call image_report(X)
    if(run%debug) call out(unit)
#endif
  end function image_param_init

  !--------------------------------------------------------------
  !
  !
  !  Core routine to setup an image
  !
  !
  logical function image_setup(X)
    type(weimag), intent(inout) :: X
    integer :: er
    logical :: zero
#ifdef DEBUG
    character(len=128), parameter :: unit='image_setup'
    if(run%debug) call in(unit)
#endif
    image_setup=F

     zero=T;if(X%from(1:3)=="his") zero=F

!    X%bites=X%esize*product(down%wsep%n(1:5))
    X%bites=X%esize*product(down%rsep%n(1:4))*down%wsep%n(5)




    down%nzb = down%rsep%n(5) / down%nzs


    if(run%mem_save<2) then
      allocate(X%bin(down%wsep%n(1),down%wsep%n(2),down%wsep%n(3),down%wsep%n(4),down%wsep%n(5)),stat=er)
      call add_mem(size(X%bin))
      if(er/=0) call seperr('cant allocate R.bin')
      X%bin=(0.0,0.0)
    else
      X%bin=>down%block
    end if

  if(run%m_nth==1 .and. zero) then
    call image_assign(X,cmplx(0.,0.))
  end if

    image_setup=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function image_setup


    !----------------------------------------------------------------
  !
  !   Setup image structure for local files if
  !     we are running MPI
  !
  !
#ifdef SEP_MPI
  logical function setup_lead_image_file(X,lead)
    type(weimag), intent(inout) :: X
    logical :: lead
    integer :: ierr,nsz
    character(len=128) :: first_lock_file,aa

      setup_lead_image_file=.false.

      !MAKE SURE THE FIRST THREAD IS A LEAD
      call MPI_Barrier(MPI_COMM_WORLD,ierr)
      nsz=len(trim(down%image_lock))
      call MPI_Bcast(nsz,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
      first_lock_file=down%image_lock
      call MPI_Bcast(first_lock_file,nsz,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)


      !FIGURE OUT WHAT THE LEAD TRACE FOR EACH MACHINE
         down%image_lock_fd=-1
         if(first_lock_file(1:nsz)/=down%image_lock(1:nsz) .or. &
           run%m_ith == 0) &
            down%image_lock_fd=mpi_sep_check_lock(down%image_lock)
         lead=.true. 
         if(-1==down%image_lock_fd) lead=.false.


      !CREATE THE LEAD FILE
      if(lead) then
        if(.not. create_image_file(X,X%tag,.false.)) then
          write(0,*) "trouble setting up mpi data"
          return
        end if
      end if
      !WAIT UNTIL THE LEAD FILES HAVE BEEN CREATED
      call MPI_Barrier(MPI_COMM_WORLD,ierr)
      !IF WE AREN'T THE LEAD FILE OPEN UP THE IMAGE FILE WITH THE RIGHT DATAPATH
      if(.not.lead) then
           call auxinout(X%tag)
       end if
      setup_lead_image_file=.true.
  end function
#endif

  logical function mpi_image_setup(X)
    type(weimag), intent(inout) :: X
    logical :: zero,create
    integer :: ierr,nsz,n
    character(len=128) :: temp
    mpi_image_setup=F

     zero=T;
     if(.not. down%adj ) zero=F



    temp=run%tag_num
    if(run%share_image) temp=run%mach

    X%tag_0=X%tag
    if(.not. make_local_tag(X%tag,x%tag_0,temp)) then
       write(0,*) "trouble making tag ",trim(X%tag)
       return
    end if

#ifdef SEP_MPI
    if(run%share_image) then
      if(.not.setup_lead_image_file(X,create)) then
        write(0,*) "trouble setting up lead image file"
        return
      end if
    else
#endif
      create=.true.
      if(.not. create_image_file(X,X%tag,.false.)) then
        write(0,*) "trouble setting up mpi data"
        return
      end if

#ifdef SEP_MPI
    end if 
#endif

    if(run%share_image) image_comm_group= create_group(create)
    image_lead_thread=create
    if(create) then
      if(.not. zero) then
        call distribute_image(X,create)
      else 
        call  image_assign(X,cmplx(0.,0.),.true.)
      end if
    end if
  
#ifdef SEP_MPI
    if(create)&
      call mpi_sep_close_lock(down%image_lock,down%image_lock_fd);
#endif
    mpi_image_setup=T
  end function





  !----------------------------------------------------------------
  !
  !
  !  Assign value to the image file
  !
  !
  !
  !
  
  subroutine image_assign(X,v,ignore_lock)
    type(weimag), intent(inout) :: X
    complex, intent(in) :: v
    integer :: izb,ierr
    logical, optional :: ignore_lock
    logical :: ignore
#ifdef DEBUG
    character(len=128), parameter :: unit='image_assign'
    if(run%debug) call in(unit)
#endif

    ignore=.false.
    if(present(ignore_lock)) ignore=ignore_lock
    X%bin=v
    do izb=1,down%nzb
       call image_write(X,izb,ignore_lock=ignore)
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_assign




  !----------------------------------------------------------------
  !
  ! Write info about image
  !
  !
  !
  subroutine image_report(X)
    type(weimag), intent(in) :: X

    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Mx: ",down%rsep%n(1), down%rsep%o(1), down%rsep%d(1)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    My: ",down%rsep%n(2), down%rsep%o(2), down%rsep%d(2)
    if(down%pcig) then
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Px: ",down%rsep%n(3), down%rsep%o(3), down%rsep%d(3)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Py: ",down%rsep%n(4), down%rsep%o(4), down%rsep%d(4)
    else
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Hx: ",down%rsep%n(3), down%rsep%o(3), down%rsep%d(3)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Hy: ",down%rsep%n(4), down%rsep%o(4), down%rsep%d(4)
    end if
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     Z: ",down%rsep%n(5),  down%rsep%o(5), &
 down%rsep%d(4)," ->",down%nzb," blocks"
    if(associated(X%bin)) then
       write(0,'(a,f9.2,a,f9.2,a)        ') "         * block: ",X%bites/1024.," / ",8*size(X%bin)/1024.," Kbytes"
       write(0,'(a,i1)')                       "         * esize=",X%esize
    end if
    write(0,*)'        ************************************************'
  end subroutine image_report

  !----------------------------------------------------------------
  !
  !
  ! Function to read image
  !
  !
  subroutine image_read(X,k,iw)
    type(weimag), intent(inout) :: X
    integer, optional :: iw
    integer, intent(in) :: k
    integer :: stat,i_w,i4,i5,i3,ierr;
    real, allocatable :: temp_buf(:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='image_read'
    if(run%debug) call in(unit)
#endif



#ifdef SEP_MPI
     if(run%share_image)&
      down%image_lock_fd= mpi_sep_open_lock(down%image_lock)
#endif
#ifdef SEP_TIMER
  if(run%m_ith==0)  call start_timer_num(timer%io_t)
#endif


    i_w=0; if(present(iw)) i_w=iw
    if(i_w==0) then 
      stat=sseek_block(X%tag,k-1,X%bites,0)
      if(X%esize==4) then
         allocate(temp_buf(size(X%bin,1),size(X%bin,2)), stat=stat)
         call add_mem(size(temp_buf))
         do i5=1,size(X%bin,5)
         do i4=1,size(X%bin,4)
         do i3=1,size(X%bin,3)
            if(size(temp_buf)*4/=sreed(trim(X%tag), temp_buf,size(temp_buf)*4)) then
              write(0,*) "trouble reading image block",i3,i4,i5,iw
              return
            end if 
            X%bin(:,:,i3,i4,i5)=cmplx(temp_buf,0.)
         end do; end do; end do;
         call del_mem(size(temp_buf))
         deallocate(temp_buf)
      else
         ierr= sreed(trim(X%tag), X%bin, X%bites)
      end if
    else
      call seperr("unimplented, varaible sampling image read")
    end if



#ifdef SEP_MPI
     if(run%share_image)&
      call mpi_sep_close_lock(down%image_lock,down%image_lock_fd)
#endif

#ifdef SEP_TIMER
  if(run%m_ith==0)  call stop_timer_num(timer%io_t)
#endif
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_read





  !----------------------------------------------------------------
  !
  !  Function to combine image
  !
  !
  !
  subroutine combine_image(X)
    type(weimag), intent(inout) :: X
    integer ::  ierr,iadd,i,it
    integer, external :: sseek
    
#ifdef SEP_MPI
    it=0

    call MPI_Barrier(MPI_COMM_WORLD,ierr)

    iadd=0;if(down%add) iadd=1;
    if(run%mpi .and. down%adj .and. image_lead_thread) then
       ierr=sseek(trim(X%tag),0,0) 
       if(run%share_image) then
         if(0/=mpi_sep_tag_sum_index(run%m_ith,X%tag,0,X%tag_0,X%bites/4,&
         down%nzb,iadd,image_comm_group))call seperr("trouble combining image");
       else
         if(run%m_ith==0) ierr= sseek(trim(x%tag_0),0,0)
         if(0/=mpi_sep_tag_sum(run%m_ith,X%tag,0,X%tag_0,X%bites/4,&
         down%nzb,iadd)) call seperr("trouble combining image ");
       end if
    end if
    call MPI_Barrier(MPI_COMM_WORLD,ierr)
     if(run%unlink) call aux_unlink(X%tag)
     call auxclose(trim(X%tag))
    if(var%use_disk_buf .and. run%unlink) call aux_unlink(X%tag_var)
#endif
  end subroutine combine_image

  !----------------------------------------------------------------
  !
  !  Distribute the image to various threads
  !
  !
  !
  subroutine distribute_image(X,use)
    type(weimag), intent(inout) :: X
    integer :: ierr,iuse,i
    logical :: use

#ifdef SEP_MPI
    if(run%mpi) then
        ierr= sseek(trim(X%tag),0,0)
       if(run%m_ith==0) then
          ierr=sseek(trim(X%tag_0),0,0)
        end if
       if(run%share_image) then
          i=1; do while(image_master(i)==0);i=i+1;end do;i=i-1;
         if(0/=mpi_sep_tag_bcast_index(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
           down%nzb,image_comm_group))call seperr("trouble distributing image ")
       else
         if(0/=mpi_sep_tag_bcast(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
           down%nzb)) call seperr("trouble distributing image ")
       end if
    end if
#endif
  end subroutine distribute_image

  integer function create_group(use)
   logical ::use
   integer :: iuse,i
   integer :: ierr
#ifdef SEP_MPI
   integer :: status(MPI_STATUS_SIZE)
#endif

   if(use) then
     iuse=1
   else 
     iuse=0
   end if

    create_group=0
#ifdef SEP_MPI
   allocate(image_master(run%m_nth))
   call add_mem(size(image_master))
   do i=1,run%m_nth 
     if(run%m_ith==0 .and. i/=1) then
        call mpi_recv(image_master(i),1,MPI_INTEGER,i-1,3242+i,MPI_COMM_WORLD,status,ierr);
     else if(i-1 == run%m_ith .and. run%m_ith/=0) then
        call mpi_send(iuse,1,MPI_INTEGER,0,3242+i,MPI_COMM_WORLD,ierr)
     end if
   end do
   if(run%m_ith==0) image_master(1)=iuse
   call  mpi_bcast(image_master,size(image_master),MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
   create_group= mpi_sep_create_group(size(image_master),image_master)
#endif

  end function





  !----------------------------------------------------------------
  !
  !
  !  Write out image
  !
  !
  !
  !
  subroutine image_write(X,k,sum,iw,nz,ignore_lock)
    logical, optional :: sum
    integer, optional :: iw,nz
    integer, intent(in) :: k
    type(weimag), intent(inout) :: X
    integer           :: i_w,n_z
    logical           :: sum_it
    logical,optional :: ignore_lock
    logical :: ignore
#ifdef DEBUG
    character(len=128), parameter :: unit='image_write'
    if(run%debug) call in(unit)
#endif
 
#ifdef SEP_TIMER
  if(run%m_ith==0)  call start_timer_num(timer%io_t)
#endif
    ignore=.false.;if(present(ignore_lock)) ignore=ignore_lock
    
#ifdef SEP_MPI
    if(run%share_image .and. .not. ignore)  &
      down%image_lock_fd= mpi_sep_open_lock(down%image_lock)
#endif
    sum_it=.false.
    if(present(sum)) sum_it=sum
    i_w=0; if(present(iw)) i_w=iw
    n_z=0; if(present(nz)) n_z=nz

    if(i_w==0) then  
      call image_write_reg(X,k,sum_it)
    else
      call image_write_var(X,k,sum_it,i_w,n_z)
    end if



#ifdef SEP_TIMER
  if(run%m_ith==0)  call stop_timer_num(timer%io_t)
#endif
#ifdef SEP_MPI
    if(run%share_image .and. .not. ignore) &
      call mpi_sep_close_lock(down%image_lock,down%image_lock_fd)
#endif

end subroutine image_write


subroutine image_write_reg(X,k,sum_it)
    type(weimag), intent(inout) :: X
    logical           :: sum_it
    integer, intent(in) :: k
   integer :: i,iz



      iz=(k-1)*size(X%bin,5)
      do i=1,size(X%bin,5)
        iz=iz+1
        call image_write_iz(X,X%bin(:,:,:,:,i),iz,sum_it)
      end do


end subroutine




subroutine image_write_var(X,k,sum_it,i_w,n_z)
    integer           :: i_w,n_z,i3,i4,ierr
    type(weimag), intent(inout) :: X
    logical           :: sum_it
    integer, intent(in) :: k
    integer :: iz,i,izlast,izfirst,ib,stat,n123
    real    :: z,zcur,ab
    logical :: write_out,stored
    complex, allocatable :: tempc_buf(:,:)
    complex, allocatable :: temp2_buf(:,:),temp3_buf(:,:,:)

    n123=size(X%bin(:,:,:,:,1))

    !Just doing a normal write
      allocate(tempc_buf(size(X%bin,1),size(X%bin,2)))
      call add_mem(size(tempc_buf))
      allocate(temp2_buf(size(X%bin,1),size(X%bin,2)))
      call add_mem(size(temp2_buf))
      allocate(temp3_buf(size(X%bin,1),size(X%bin,2),size(X%bin,5)))
      call add_mem(size(temp3_buf))

      !
      !
      !First check to see  what we can write
      !
      !

      izlast=min(size(var%z_samp(i_w)%dz),size(X%bin,5)*k) !last iz step done
      zcur=var%z_samp(i_w)%zloc(izlast) ! depth we have gone done to




      z=down%rsep%o(5);iz=1;
      do while(z<=zcur .and. iz <= down%rsep%n(5)) 
        if(.not. var%filled(iz)) then
          write_out=.true.
          do i=1,var%z_samp(i_w)%npts
            if(var%z_samp(i_w)%iloc(i,iz) > izlast .and.  &
               abs(var%z_samp(i_w)%frac(i,iz)) > .001) &
                write_out=.false.
          end do 
          if(write_out) then
            stat=sseek_block(X%tag,iz-1,X%bites/size(X%bin,5),0)
            do i4=ihy_beg,ihy_end
              do i3=ihx_beg,ihx_end
                tempc_buf=0.
                temp3_buf=X%bin(:,:,i3,i4,:)
                 ab=maxval(real(temp3_buf))
                call lagrange_interp(.true.,i_w,k,iz,temp3_buf,tempc_buf,&
                   temp2_buf,X%tag_var,i3,i4)
                call write_mx_my(X%tag,X%esize,size(X%bin,1),size(X%bin,2)&
                  ,tempc_buf,sum_it)
              end do
            end do
            var%filled(iz)=.true.
          end if 
         end if
         z=z+down%rsep%d(5);iz=iz+1
       end do

      !
      !
      ! Check to see if we can clear any incore or disk memory
      !
      !

      ! Find the smallest iz still needed
      izfirst=izlast
      do iz=1, size(var%filled) 
        if(.not. var%filled(iz)) then
          do i=1,var%z_samp(i_w)%npts
            if(abs(var%z_samp(i_w)%frac(i,iz))>.001) &
              izfirst=min(izfirst,var%z_samp(i_w)%iloc(i,iz)) 
          end do
        end if 
      end do 

      do i=1,var%ndisk_buf
        if(var%disk_buf(i) < izfirst ) var%disk_buf(i)=0
      end do
    
      do i=1,var%nmem_buf
        if(var%mem_buf(i) < izfirst ) var%mem_buf(i)=0
      end do
    
      ! 
      ! 
      !store any newly created image components that are still needed
      ! 
      ! 

      iz=(k-1)*size(X%bin,5)
      do i=1,n_z
        iz=iz+1
        if(iz>=izfirst) then !we need to store this image compoenent
          stored=.false.;ib=1;
          do while(ib <= var%nmem_buf .and. .not. stored) 
            if(var%mem_buf(ib)==0) then
              var%v_i_buf(:,:,:,:,ib)=X%bin(:,:,:,:,i)
              stored=.true.; var%mem_buf(ib)=iz
              var%z_samp(i_w)%status(iz)=ib
            end if 
            ib=ib+1; 
          end do
          ib=1;
          do while(ib <= var%ndisk_buf .and. .not. stored) 
            if(var%disk_buf(ib)==0) then
              stat=sseek_block(X%tag_var,ib-1,n123*8,0)
              ierr= srite(X%tag_var, X%bin(:,:,:,:,i), n123*8)
              stored=.true.; var%disk_buf(ib)=iz
              var%z_samp(i_w)%status(iz)=-ib
            end if 
            ib=ib+1;
          end do
        end if 
      end do
      call del_mem(size(temp2_buf))
      call del_mem(size(temp3_buf))
      call del_mem(size(tempc_buf))
      deallocate(tempc_buf,temp2_buf,temp3_buf)


end subroutine

  subroutine image_write_iz(X,image,k,sum_it)
    logical           :: sum_it
    type(weimag), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat,i3,i4,blk,rem
    complex :: image(:,:,:,:)
    
#ifdef DEBUG
    character(len=128), parameter :: unit='image_write_iz'
    if(run%debug) call in(unit)
#endif
#ifdef SEP_TIMER
  if(run%m_ith==0)  call start_timer_num(timer%io_t)
#endif
    
       stat=sseek_block(X%tag,k-1,X%bites/size(X%bin,5),0)
       do i4=ihy_beg,ihy_end
         do i3=ihx_beg,ihx_end
             call write_mx_my(X%tag,X%esize,size(image,1),size(image,2),image(:,:,i3,i4),sum_it)
          end do
       end do
!    call to_history("zblock",k,X%tag)
    
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine image_write_iz


  subroutine write_mx_my(tag,esize,n1,n2,image,sum_it)
    logical :: sum_it
    character(len=*) :: tag
    integer :: n1,n2,ierr,esize,blk,rem
    complex :: image(:,:),tempc(n1,n2)
    real :: tempr(n1,n2)
     
    if(esize==4) then
      if(sum_it) then
        if(esize*n1*n2/=sreed(tag,tempr,n1*n2*4)) &
          call erexit("trouble reading in image 4")
        tempr=tempr+real(image) 
        ierr=sseek(tag,-n1*n2*4,1)
      else 
        tempr=real(image)
      end if
      if(esize*n1*n2/=srite(tag,tempr,n1*n2*4)) &
        call erexit("trouble writing out image 4")
    else
      if(sum_it) then
        if(esize*n1*n2/=sreed(tag,tempc,n1*n2*8))  then
          write(0,*) "attempting to write",size(tempc)," tag=",trim(tag),run%m_ith,"=thread"
          call erexit("trouble reading in image 8a")
        end if
        tempc=tempc+image
        ierr=sseek(tag,-n1*n2*8,1)
      else
        tempc=image
      end if
      if(8*n1*n2/=srite(tag,tempc,n1*n2*8)) then
         write(0,*) "attempting to write",size(tempc)," tag=",trim(tag),run%m_ith,"=thread"
        call erexit("trouble writing out image 8b" )
      end if
    end if
  end subroutine 


  !----------------------------------------------------------------
end module wei_obj_R

