module wei_obj_S

  use wei_run
  use wei_down_type
  use wei_var
  use wei_var_samp
  use wei_type
  !  use wei_work
  use wei_xmap
  !---------------
  implicit none

  real,    dimension(:), pointer, private ::  sloline
  real,    dimension(:,:), pointer, private,save ::  fx_s,fy_s
  integer,    dimension(:,:), pointer, private,save ::  ix_s,iy_s
  integer,    dimension(:,:), pointer, private,save ::  jx_s,jy_s
  real,    dimension(:,:), pointer, private,save ::  fx_g,fy_g
  integer,    dimension(:,:), pointer, private,save ::  ix_g,iy_g
  integer,    dimension(:,:), pointer, private,save ::  jx_g,jy_g
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  type(weslow),    public :: S   !Slowness
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  type(weslowmap), public :: slo
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!

contains
  !----------------------------------------------------------------
  !INITIALIZE THE SLOWNESS FUNCTION
  ! GIVEN UP TO TWO TAGS DESCRIBING THE SLOWNESS
  logical function wei_slow_init(X,tag_S,tag_G)
    type(weslow),       intent(inout) :: X
    character(len=128), intent(in)    :: tag_S,tag_G
    real :: pos,fract,pos2
    integer :: i1,i2
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='wei_slow_init' 
    if(run%debug) call in(unit)
#endif

    wei_slow_init=F 


    if(down%conv_wave) then
       X%s_s=X%s1
       X%s_g=X%s2
       X%s_s%tag = tag_S
       X%s_g%tag = tag_g
       if(.not. slowness_history_init(X%s_s,down%ssep)) then
          write(0,*) "trouble init slowness history"
          return
       end if
       if(.not. slowness_history_init(X%s_g,down%s2sep)) then
          write(0,*) "trouble init slowness history"
          return
       end if
    else
       X%s_s=X%s1
       X%s_g=X%s1
       X%s_s%tag = tag_S
       if(.not. slowness_history_init(X%s_s,down%ssep)) then
          write(0,*) "trouble init slowness history"
          return
       end if
    end if

    !allocate array
    if(run%mem_save<1) then
       allocate(X%pba(down%wsep%n(1),down%wsep%n(2),down%wsep%n(3),down%wsep%n(4),down%nzs),stat=er)
       call add_mem(size(X%pba))
       if(er/=0) call seperr('cant allocate S.pba')
       X%pba=(0.0,0.0)
    else
       nullify(X%pba)
       allocate(X%pba_s(down%wsep%n(1),down%wsep%n(2)),stat=er)
       call add_mem(size(X%pba_s))
    end if
       allocate(jx_s(down%wsep%n(1),down%wsep%n(3)),stat=er)
       allocate(jy_s(down%wsep%n(2),down%wsep%n(4)),stat=er)
       allocate(fx_s(down%wsep%n(1),down%wsep%n(3)),stat=er)
       allocate(fy_s(down%wsep%n(2),down%wsep%n(4)),stat=er)
       allocate(ix_s(down%wsep%n(1),down%wsep%n(3)),stat=er)
       allocate(iy_s(down%wsep%n(2),down%wsep%n(4)),stat=er)
       allocate(jx_g(down%wsep%n(1),down%wsep%n(3)),stat=er)
       allocate(jy_g(down%wsep%n(2),down%wsep%n(4)),stat=er)
       allocate(fx_g(down%wsep%n(1),down%wsep%n(3)),stat=er)
       allocate(fy_g(down%wsep%n(2),down%wsep%n(4)),stat=er)
       allocate(ix_g(down%wsep%n(1),down%wsep%n(3)),stat=er)
       allocate(iy_g(down%wsep%n(2),down%wsep%n(4)),stat=er)
       call add_mem(2*size(fx_s)+2*size(fy_s))
       do i2=1,down%wsep%n(3)
         do i1=1,down%wsep%n(1)
           fract=(xmap%sx(i1,i2)-down%ssep%o(1))/down%ssep%d(1)+1.
           call figure_index(fract,down%ssep%n(1),ix_s(i1,i2),jx_s(i1,i2),fx_s(i1,i2))
           fract=(xmap%rx(i1,i2)-down%ssep%o(1))/down%ssep%d(1)+1.
           call figure_index(fract,down%ssep%n(1),ix_g(i1,i2),jx_g(i1,i2),fx_g(i1,i2))
         end do
       end do
       do i2=1,down%wsep%n(4)
         do i1=1,down%wsep%n(2)
           fract=(xmap%sy(i1,i2)-down%ssep%o(2))/down%ssep%d(2)+1.
           call figure_index(fract,down%ssep%n(2),iy_s(i1,i2),jy_s(i1,i2),fy_s(i1,i2))
           fract=(xmap%ry(i1,i2)-down%ssep%o(2))/down%ssep%d(2)+1.
           call figure_index(fract,down%ssep%n(2),iy_g(i1,i2),jy_g(i1,i2),fy_g(i1,i2))
         end do
       end do




    call find_max_slow()

    down%S=S
    wei_slow_init=T 

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function wei_slow_init

   subroutine figure_index(fract,n,i,j,f)
      real :: fract,f
      integer :: i,j,n,index
        if(n==1) then
           i=1;f=1.; j=1;
        else
           i=fract
           if(i<=0) then
             i=1; j=2;f=0.
           else if (i >=n) then
             i=n-1; j=n;f=1.
           else
             j=i+1; f=fract-i
           end if 
       end if
   end subroutine


  !FIND THE MAXIMUM SLOWNESS AT EACH DEPTHS
  subroutine find_max_slow()
    integer :: i,ndone,block,j,ierr
    character(len=128) :: tag_1,tag_2
    real,allocatable :: junk(:)
    ndone=0;
    allocate(junk(S%s_s%struct%n(3)))
       call add_mem(size(junk))
    allocate(down%max_slow(S%s_s%struct%n(3)))
       call add_mem(size(down%max_slow))
    i=0
    do while(ndone< S%s_s%struct%n(3))
       i=i+1
       block=min(s%s_s%struct%n(3)-ndone,size(S%s_s%bin,3))
       call slowness_read(S,i,block)
       if(down%conv_wave) then
          do j=1,block
             down%max_slow(j+ndone)=max(maxval(real(S%s_s%bin(:,:,j))),maxval(real(S%s_g%bin(:,:,j))))
          end do
       else
          do j=1,block
             down%max_slow(j+ndone)=maxval(real(S%s_s%bin(:,:,j)))
          end do
       end if
       ndone=ndone+block
    end do
    call del_mem(size(junk))
    deallocate(junk)
  end subroutine find_max_slow



  !----------------------------------------------------------------
  !
  !INITIALIZE SLOWNESS FROM A FILE
  !
  logical function slowness_history_init(X,struct)
    type(weslow_comp), intent(inout) :: X
    type(sep3d) :: struct
    integer  :: stat
    character(len=128) :: temp
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_history_init' 
    if(run%debug) call in(unit)
#endif

    slowness_history_init=F
    call auxinout(X%tag)
    X%from = "history"
    X%struct=struct

    call from_param("velocity",X%velocity,.false.)
    call from_aux(trim(X%tag),"esize",X%esize,4)

    if(.not. slowness_setup(X)) then
       write(0,*) "trouble with slowness setup"
       return
    end if

    if(down%order(1:3)=="VAR") then
      if(.not. make_local_tag(X%tag_var,"VS_",run%tag_num)) then
          write(0,*) "trouble creating local tag"
          return
       end if
       call auxscr(X%tag_var)
       stat= srite(X%tag_var,0.,4)
    end if
    slowness_history_init=T

#ifdef DEBUG
    if(run%debug) call slowness_report(X)
    if(run%debug) call out(unit)
#endif
  end function slowness_history_init
  !----------------------------------------------------------------
  ! SETUP SLOWNESS
  ! 
  logical function slowness_setup(X)
    type(weslow_comp), intent(inout) :: X
    integer :: er,i1,i2,i3
    character(len=128) :: temp
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_setup' 
    if(run%debug) call in(unit)
#endif

    slowness_setup=F

    X%bites = X%esize * product( (/X%struct%n(1),X%struct%n(2),down%nzs/) )
#ifdef SEP_MPI
    if(X%tag_0(1:1)=="/") then
       temp=""
    else
       temp=trim(run%local_path)
    end if

    if(run%mpi) then
       X%tag_0=X%tag;
       if(.not. make_local_tag(X%tag_local,X%tag_0,run%tag_num)) then
         write(0,*) "trouble making local tag"
         return
       end if
       x%tag=X%tag_local
       call auxinout(trim(X%tag))
       !       call auxscr(trim(X%tag))
       call sep3d_write_description(trim(X%tag),X%struct)
    end if
    call distribute_slow(X)
#endif


    down%nzb = down%rsep%n(3) / down%nzs
    !!--------------------
    allocate(X%bin(X%struct%n(1),X%struct%n(2),down%nzs),stat=er)
    call add_mem(size(x%bin))
    if(er/=0) call seperr('cant allocate S.bin')


    do i3=1,size(X%bin,3)
       do i2=1,size(X%bin,2)
          do i1=1,size(X%bin,1)
             X%bin(i1,i2,i3)=(0.0,0.0)
          end do
       end do
    end do
    !!--------------------
    !! slowness interpolation map
    allocate( sloline(down%nzs),stat=er)
    call add_mem(size(sloline))
    if(er/=0) call seperr('cant allocate  sloline')
    slowness_setup=T

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function slowness_setup



  !----------------------------------------------------------------
  !WRITE INFO ABOUT SLOWNESS
  subroutine slowness_report(X)
    type(weslow_comp), intent(in) :: X

    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Mx: ",X%struct%n(1), X%struct%o(1), X%struct%d(1)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    My: ",X%struct%n(2), X%struct%o(2), X%struct%d(2)
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     Z: ",X%struct%n(3),  X%struct%o(3),  X%struct%d(3)," ->",down%nzb," blocks"
    if(X%velocity) then
       write(0,'(a,i1,a)')               "         *    esize=",X%esize," velocity"
    else 
       write(0,'(a,i1,a)')               "         *    esize=",X%esize," slowness"
    end if
    if(associated(X%bin)) then
       write(0,'(a,f6.2,a,f6.2,a)        ') "         * block: ",X%bites/1024.," / ",8*size(X%bin)/1024.," Kbytes"
       write(0,*)'        ************************************************'
    end if

  end subroutine slowness_report
  !--------------------------------------------------------------
  ! READ SLOWNESS
  subroutine slowness_read(X,k,nzs)
    type(weslow), intent(inout) :: X
    integer, intent(in) :: k
    integer, optional :: nzs
    integer :: stat,nz,i
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_read' 
    if(run%debug) call in(unit)
#endif


    call slowness_read_comp(X%s_s,k,nzs)
    if(down%conv_wave) then
       call slowness_read_comp(X%s_g,k,nzs)
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine slowness_read

  !READ A SLOWNESS COMPONENT
  subroutine slowness_read_comp(X,k,nzs)
    type(weslow_comp), intent(inout) :: X
    integer, intent(in) :: k
    integer, optional :: nzs
    integer :: stat,nz,i
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_read_comp' 
    if(run%debug) call in(unit)
#endif

    nz=size(X%bin,3); if(present(nzs)) nz=nzs

    stat=sseek_block(X%tag,k-1,X%bites,0)
       stat= sreed(X%tag, X%bin, X%bites/size(X%bin,3)*nz)
       if(stat/=X%bites/size(X%bin,3)*nz) then
         write(0,*) "Tried to readin ",X%bites/size(X%bin,3)*nz, " read=",stat,trim(X%tag) 
          call seperr("error reading in slowness")
       end if
       if(minval(X%bin(:,:,:nz)) < .001) then
           write(0,*) "MINIMUM VELOCITY",run%m_ith,k,minval(X%bin(:,:,:nz)),trim(X%tag)
           call seperr("velocity 0 or less")
       end if
       if(X%velocity) X%bin=1./X%bin

    if(nz <  size(X%bin,3)) then
       do i=nz,size(X%bin,3)
          X%bin(:,:,i)=X%bin(:,:,nz)
       end do
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine slowness_read_comp

  !--------------------------------------------------------------
  !DISTRIBUTE SLOWNESS TO ALL THE THREADS
  subroutine distribute_slow(X)
    integer :: stat
    type(weslow_comp), intent(inout) :: X

#ifdef SEP_MPI
    if(run%mpi) then
       stat= sseek(trim(X%tag),0,0)
       if(run%m_ith==0) then
          stat= sseek(trim(X%tag_0),0,0);
       end if
       if(0/=mpi_sep_tag_bcast(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
       X%struct%n(3)/down%nzs)) call seperr("trouble distributing slow ")
    end if
#endif
  end subroutine distribute_slow
  !--------------------------------------------------------------
  !----------------------------------------------------------------
  ! WRITE OUT SLOWNESS
  subroutine slowness_write(X,k)
    type(weslow_comp), intent(inout) :: X
    integer :: k
    integer :: stat
    integer :: iz
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_write'
    if(run%debug) call in(unit)
#endif

    stat=sseek_block(X%tag,k-1,X%bites,0)
    stat= srite(X%tag, X%bin, X%bites)

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine slowness_write
  !----------------------------------------------------------------



  !ALLOCATE SPACE FOR SLOWNESS
  subroutine weislow_allocate()
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weislow_allocate' 
    if(run%debug) call in(unit)
#endif

    !! FKop slowness arrays
    allocate(slo%sz_r(down%nzs,down%nref+1),stat=er)
       call add_mem(size(slo%sz_r))
    if(er/=0) call seperr('cant allocate slo%sz_r')
    allocate(slo%sz_s(down%nzs,down%nref+1),stat=er)
       call add_mem(size(slo%sz_s))
    if(er/=0) call seperr('cant allocate slo%sz_s')

    allocate(slo%th(run%o_nth))
       call add_mem(size(slo%th))

    slo%th(1)%sz=>slo%sz_s

    allocate(slo%nrs_s(down%nzs),stat=er)
       call add_mem(size(slo%nrs_s))
    if(er/=0) call seperr('cant allocate slo%nrs_nzss')
    allocate(slo%nrs_r(down%nzs),stat=er)
       call add_mem(size(slo%nrs_r))
    if(er/=0) call seperr('cant allocate slo%nrs_g')
    slo%th(1)%nrs=>slo%nrs_s

    allocate(slo%bound_r(down%nzs,0:down%nref+1),stat=er)
       call add_mem(size(slo%bound_r))
    if(er/=0) call seperr('cant allocate slo%bound_r')
    allocate(slo%bound_s(down%nzs,0:down%nref+1),stat=er)
       call add_mem(size(slo%bound_s))
    if(er/=0) call seperr('cant allocate slo%bound_s')
    slo%bound_r=0.
    slo%bound_s=0.
    slo%th(1)%bound=>slo%bound_s


    !! FXop interpolation array
    if(run%mem_save<1) then
       allocate(slo%si_r(down%wsep%n(1),down%wsep%n(2), down%wsep%n(3),down%wsep%n(4), down%nzs),stat=er)   
       call add_mem(size(slo%si_r)/2)
       allocate(slo%si_s(down%wsep%n(1),down%wsep%n(2), down%wsep%n(3),down%wsep%n(4), down%nzs),stat=er)   
       call add_mem(size(slo%si_s)/2)
       slo%th(1)%si=>slo%si_s
       if(er/=0) call seperr('cant allocate slo%si')
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weislow_allocate
  !----------------------------------------------------------------
  !FIND THE MEAN VALUE OF THE SLOWNESS
  !
  real function slowness_mean(X,k) result(smean)
    type(weslow), intent(in) :: X
    integer :: k
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_mean'
    if(run%debug) call in(unit)
#endif DEBUG


    if(down%conv_wave) then
       smean = (sum(real(X%s_s%bin(:,:,k)))+sum(real(X%s_g%bin(:,:,k))))/(2* X%s_s%struct%n(1) * X%s_s%struct%n(2))
    else
       smean = sum(real(X%s_s%bin(:,:,k)))/(X%s_s%struct%n(1) * X%s_s%struct%n(2))
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif DEBUG
  end function slowness_mean
  !----------------------------------------------------------------
  !FIND THE MAX VALUE OF THE SLOWNESS
  !
  real function slowness_max(X,k) result(smax)
    type(weslow), intent(in) :: X
    integer :: k
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_max'
    if(run%debug) call in(unit)
#endif DEBUG

    if(down%conv_wave) then
       smax =max(maxval(real(X%s_s%bin(:,:,k))),maxval(real(X%s_g%bin(:,:,k))))
    else
       smax = maxval(real(X%s_s%bin(:,:,k)))
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif DEBUG
  end function slowness_max
  !----------------------------------------------------------------
  !FIND THE MIN VALUE OF THE SLOWNESS
  !
  real function slowness_min(X,k) result(smin)
    type(weslow), intent(in) :: X
    integer :: k
#ifdef DEBUG
    character(len=128), parameter :: unit='slowness_min'
    if(run%debug) call in(unit)
#endif DEBUG

    if(down%conv_wave) then
       smin =min(minval(real(X%s_s%bin(:,:,k))),minval(real(X%s_g%bin(:,:,k))))
    else
       smin = minval(real(X%s_s%bin(:,:,k)))
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif DEBUG
  end function slowness_min

  !----------------------------------------------------------------
  ! FOR A GIVEN MIDPOINT FIND THE SLOWNESS AT EACH OFFSET
  ! A MEASURE TO REDUCE STORAGE
  subroutine fill_big_pba(X)
    type(weslow), intent(inout) :: X
    integer :: ihx,ihy,imx,imy,si_y,sj_y,ith,izs,gi_y,gj_y
    real    :: xc,yc,hx,hy,sf_y,gf_y
    real,pointer :: cubes(:,:,:),cubeg(:,:,:)
    real,pointer :: fx(:,:),fy(:,:)
    integer,pointer :: ix(:,:),jx(:,:),iy(:,:),jy(:,:)
    logical :: one
#ifdef DEBUG
    character(len=128), parameter :: unit='fill_big_pba'
    if(run%debug) call in(unit)
#endif DEBUG
   if(run%mem_save==0) then

    cubes=>S%s_s%bin(:,:,:)
    cubeg=>S%s_s%bin(:,:,:)
    if(down%conv_wave) cubeg=>X%s_g%bin(:,:,:)
     
#ifdef SEP_TIMER
    call start_timer_num(timer%s_dis_t)
#endif

    one=.false.
    if(size(cubes,3)==1) one=.true.

   do ihy=1,down%wsep%n(4)
    do ihx=1,down%wsep%n(3)
      do imy=1,down%wsep%n(2)
         si_y=iy_s(imy,ihy) 
         sj_y=jy_s(imy,ihy) 
         sf_y=fy_s(imy,ihy) 
         gi_y=iy_g(imy,ihy) 
         gj_y=jy_g(imy,ihy) 
         gf_y=fy_g(imy,ihy) 
         if(one) then
           do imx=1,down%wsep%n(1)
             s%pba(imx,imy,ihx,ihy,1)=cmplx(&
              (1.-sf_y)*(1.-fx_s(imx,ihx))*cubes(ix_s(imx,ihx),si_y,1)+&
              (1.-sf_y)*(fx_s(imx,ihx))*cubes(jx_s(imx,ihx),si_y,1)+&
              (sf_y)*(1.-fx_s(imx,ihx))*cubes(ix_s(imx,ihx),sj_y,1)+&
              (sf_y)*(fx_s(imx,ihx))*cubes(jx_s(imx,ihx),sj_y,1),&
              (1.-gf_y)*(1.-fx_g(imx,ihx))*cubeg(ix_g(imx,ihx),gi_y,1)+&
              (1.-gf_y)*(fx_g(imx,ihx))*cubeg(jx_g(imx,ihx),gi_y,1)+&
              (gf_y)*(1.-fx_g(imx,ihx))*cubeg(ix_g(imx,ihx),gj_y,1)+&
              (gf_y)*(fx_g(imx,ihx))*cubeg(jx_g(imx,ihx),gj_y,1) &
              )
           end do
         else 
           do imx=1,down%wsep%n(1)
             s%pba(imx,imy,ihx,ihy,:)=cmplx(&
              (1.-sf_y)*(1.-fx_s(imx,ihx))*cubes(ix_s(imx,ihx),si_y,:)+&
              (1.-sf_y)*(fx_s(imx,ihx))*cubes(jx_s(imx,ihx),si_y,:)+&
              (sf_y)*(1.-fx_s(imx,ihx))*cubes(ix_s(imx,ihx),sj_y,:)+&
              (sf_y)*(fx_s(imx,ihx))*cubes(jx_s(imx,ihx),sj_y,:),&
              (1.-gf_y)*(1.-fx_g(imx,ihx))*cubeg(ix_g(imx,ihx),gi_y,:)+&
              (1.-gf_y)*(fx_g(imx,ihx))*cubeg(jx_g(imx,ihx),gi_y,:)+&
              (gf_y)*(1.-fx_g(imx,ihx))*cubeg(ix_g(imx,ihx),gj_y,:)+&
              (gf_y)*(fx_g(imx,ihx))*cubeg(jx_g(imx,ihx),gj_y,:) &
              )

           end do
         end if
      end do
    end do
   end do

end if


#ifdef SEP_TIMER
    call stop_timer_num(timer%s_dis_t)
#endif
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif DEBUG
 end subroutine fill_big_pba


  !----------------------------------------------------------------
  ! FOR A GIVEN MIDPOINT FIND THE SLOWNESS AT EACH OFFSET
  ! A MEASURE TO REDUCE STORAGE
  subroutine fill_small_pba(ihx,ihy,ith,izs)
    integer :: ihx,ihy,imx,imy,i_y,j_y,ith,izs
    real    :: xc,yc,hx,hy,f_y
    real,pointer :: cube(:,:)
    real,pointer :: fx(:,:),fy(:,:)
    integer,pointer :: ix(:,:),jx(:,:),iy(:,:),jy(:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='fill_small_pba'
    if(run%debug) call in(unit)
#endif DEBUG

    cube=>s%s_s%bin(:,:,izs)
    if( slo%th(ith)%source) then
      fx=>fx_s
      ix=>ix_s
      jx=>jx_s
      fy=>fy_s
      iy=>iy_s
      jy=>jy_s
    else
      fx=>fx_g
      ix=>ix_g
      jx=>jx_g
      fy=>fy_g
      iy=>iy_g
      jy=>jy_g
      if(down%conv_wave) cube=>s%s_g%bin(:,:,izs)
    end if
     


#ifdef SEP_TIMER
    call start_timer_num(timer%s_dis_t)
#endif

     if(slo%th(ith)%source) then
      do imy=1,down%wsep%n(2)
         i_y=iy(imy,ihy) 
         j_y=jy(imy,ihy) 
         f_y=fy(imy,ihy) 
         do imx=1,down%wsep%n(1)
           s%pba_s(imx,imy)=&
            (1.-f_y)*(1.-fx(imx,ihx))*cube(ix(imx,ihx),i_y)+&
            (1.-f_y)*(fx(imx,ihx))*cube(jx(imx,ihx),i_y)+&
            (f_y)*(1.-fx(imx,ihx))*cube(ix(imx,ihx),j_y)+&
            (f_y)*(fx(imx,ihx))*cube(jx(imx,ihx),j_y)
         end do
      end do
    else
      do imy=1,down%wsep%n(2)
         i_y=iy(imy,ihy) 
         j_y=jy(imy,ihy) 
         f_y=fy(imy,ihy) 
         do imx=1,down%wsep%n(1)
           s%pba_s(imx,imy)=&
            (1.-f_y)*(1.-fx(imx,ihx))*cube(ix(imx,ihx),i_y)+&
            (1.-f_y)*(fx(imx,ihx))*cube(jx(imx,ihx),i_y)+&
            (f_y)*(1.-fx(imx,ihx))*cube(ix(imx,ihx),j_y)+&
            (f_y)*(fx(imx,ihx))*cube(jx(imx,ihx),j_y)
         end do
      end do
    end if


#ifdef SEP_TIMER
    call stop_timer_num(timer%s_dis_t)
#endif
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif DEBUG
 end subroutine fill_small_pba


  !----------------------------------------------------------------
  !! push slowness from bin to ss and sr
  subroutine partition_slow(win,wout,ifk,izs,ith)
    complex :: win(:,:,:,:)
    complex :: wout(:,:,:,:)
    integer :: imx,imy,ihx,ihy,ifk,i,izs,ith
    real    :: xc,yc,avg
    real :: bound(0:down%nref+1)
    logical :: source
#ifdef DEBUG
    character(len=128), parameter :: unit='partition_slow' 
    if(run%debug) call in(unit)
#endif


   bound=slo%th(ith)%bound(izs,:)


    do ihy=1,down%wsep%n(4)
       do ihx=1,down%wsep%n(3)
          call fill_small_pba(ihx,ihy,ith,izs)
          do imy=1,down%wsep%n(2)
             do imx=1,down%wsep%n(1)
                   avg=S%pba_s(imx,imy)
                   if(avg <= bound(ifk-1) .or. avg > bound(ifk)) then  
                     wout(imx,imy,ihx,ihy)=cmplx(0.,0.)
                   else
                     wout(imx,imy,ihx,ihy)=win(imx,imy,ihx,ihy)
                   end if
             end do
          end do
       end do
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine partition_slow
  subroutine var_slow(iwb)
    integer :: iwb
    call create_modified_slowness(S%s_s%tag_local,S%s_s%tag_var,s%s_s%velocity, &
    S%s_s%struct%n,s%s_s%struct%o,S%s_s%struct%d, &
    var%z_samp(iwb)%dz,var%z_samp(iwb)%zloc) 
    s%s_s%tag=s%s_s%tag_var
    if(down%conv_wave) then
       call create_modified_slowness(S%s_g%tag_local,S%s_g%tag_var,s%s_g%velocity, &
       S%s_g%struct%n,s%s_g%struct%o,S%s_g%struct%d, &
       var%z_samp(iwb)%dz,var%z_samp(iwb)%zloc) 
       s%s_g%tag=s%s_g%tag_var
    end if
  end subroutine var_slow
  !----------------------------------------------------------------
  subroutine sellect_slow_comp(source,ith)
    logical :: source 
    integer :: ith

    slo%th(ith)%source=source
    if(source) then
       slo%th(ith)%si=>slo%si_s
       slo%th(ith)%nrs=>slo%nrs_s
       slo%th(ith)%bound=>slo%bound_s
       slo%th(ith)%sz=>slo%sz_s
    else
       slo%th(ith)%si=>slo%si_r
       slo%th(ith)%nrs=>slo%nrs_r
       slo%th(ith)%bound=>slo%bound_r
       slo%th(ith)%sz=>slo%sz_r
    end if

  end subroutine sellect_slow_comp


  !----------------------------------------------------------------
end module wei_obj_S
