module wei_down
  use sep
  use wei_var
  use wei_run
  use wei_down_type
  integer,private,save :: idone,iblocks,nblocks,nzb,nwb,dprint !Number of blocks for this thread
  real, private, save :: nprint
  complex, pointer :: tmp_on(:,:,:,:,:)
  !----------------------------------------------------------------



contains
  !----------------------------------------------------------------

  !INITIALIZATION FUNCTION FOR DOWNWARD CONTINUATION
  logical function down_init()
    logical :: tt
    integer :: er,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='process_init'
    call in(unit)
#endif

    down_init=F
    !OPERATION SETUP
    call from_param('adj', down%adj, .true.)
    call from_param('add', down%add, .false.)


    !! inversion begins with adjoint, too
    if(run%operation(1:10) .eq. 'inversion.') then
       down%adj=.true. 
    end if

    !HANDLE CONVERTED WAVE 
    ierr=getch("S_S","d",ierr)+getch("S_G","d",ierr)
    if(2==ierr) then
        down%VS='S_S'   !!       slowness
        down%VG='S_G'   !!       slowness
        down%conv_wave=.true.
    else if( 1==ierr) then
       write(0,*) "Must provide S_S and S_G"
       return
    else
        down%VS='S'
        down%VG='S'
        down%conv_wave=.false.
    end if

    call from_param("shot_profile",down%shot_profile,.false.)
    if(down%shot_profile) call seperr("shot profile unimplemented")



    !HANDLE SHARED IMAGE FILE
    if(run%share_image) then
       down%image_lock="LOCK."//trim(run%mach)
#ifdef SEP_MPI
       call mpi_sep_close_lock(down%image_lock,-1)
#endif
    end if
    

    down%dS='dS' !! delta slowness
    down%D='D'   !!       data
    down%U='U'   !!       downward continued wavefield
    down%R='R'   !!       image
    down%dR='dR' !! delta image
    down%W='W'   !!       wavefield


    !NUMBER OF STEPS IN Z AND W TO HANDLE AT A TIME
    call from_param("nzs",down%nzs,1); down%nzb=0;
    call from_param("nws",down%nws,1); down%nwb=0;


    !Maximum velocity 
    call from_param("maxslow",down%maxslow,.false.)

    !!FIGURE OUT VARIABLE SLOWNESS
    call from_param("order",down%order,"ANY")
    if(.not. wei_var_init()) then
      write(0,*) "trouble initializing variable step"
     end if
    if(down%order(1:3) == "VAR") then
    else if(down%order(1:3) == "ANY") then
       down%order="OZ";if(down%nzs > 10) down%order="OF"
    else if(down%order(1:2) == "OF" .or. down%order(1:2) == "OZ" ) then
     else 
     write(0,*) "invalid order ",trim(down%order), " must be OF,OZ,VAR, or ANY"
     return
    end if

    if(run%mem_save > 1) down%order="OF"
  
    !! read in axes info from the parameter file
    if(.not. process_parameters()) then
      write(0,*) "trouble initializing parameters"
      return
    end if


    !IF RUNNING IN MEMORY SAVE WE ARE GOING TO REUSE MEMORY BLOCK
    if(run%mem_save > 1) then
      if(down%nws >1  .or. down%nzs>1) then
        write(0,*) "memory_save >1  option only available with nws=1 and nzs=1"
        return
      end if
      allocate(down%block(down%wsep%n(1),down%wsep%n(2),down%wsep%n(3),down%wsep%n(4),1))
      call add_mem(size(down%block))
    end if


    !! number of reference velocities
    call from_param("nref",down%nref,1)

    !! CIG in offset or offset ray parameter
    call from_param("pCIG",down%pCIG,.false.)

    if(.not. param_check()) then 
#ifdef SEP_MPI
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
#endif
       call seperr( "Parameter matching problem")
    end if


    if(run%verb) call down_report()

    call from_param("verb",tt,.false.)
    dprint=1000
    if(tt) then
       nprint=-.000001;
       dprint=10;
       if(run%m_ith==0) dprint=.0001;
    else
       nprint=1000;
    end if
 

 


    down_init=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function down_init

  subroutine allocate_tmps()
   integer :: er,nbig
   integer :: ncopies

   ncopies=2*run%o_nth
   if(run%fk_extra <0 .or. run%mem_save > 4) &
    call seperr("fk_extra must be between 0 and 4")
   if(run%fk_extra >0 .and. run%mem_save >2) &
     call seperr("can't handlke mem_save >2 and fk_extra >0")
   if(run%mem_save==2) ncopies=1
  
   nbig=max(ncopies+run%fk_extra*run%o_nth,var%nmem_buf)
   if(run%mem_save < 3) then
     !! allocate tmp arrays for interpolation of multiple fk ops
     allocate(tmp_on(down%Wsep%n(1),down%wsep%n(2),down%wsep%n(3),down%wsep%n(4),nbig),stat=er)
     call add_mem(size(tmp_on))
     if(er/=0) call seperr('cant allocate tmp1_on')
   end if

    

  end  subroutine


  !----------------------------------------------------------------
  logical function param_check()
    real ::hext(2),mext(2),sext(2)
    logical :: image_real,only_zero_offset
    param_check=.true.
    
    if(wei_dfile .and. wei_rfile .and. .not. down%shot_profile ) then
      call axis_check_file("amx",1,down%Dsep,down%Rsep,.false.,param_check)
      call axis_check_file("amy",2,down%Dsep,down%Rsep,.false.,param_check)
      if(.not. down%pCig) then
        call axis_check_file("ahx",3,down%Dsep,down%Rsep,.false.,param_check)
        call axis_check_file("ahy",4,down%Dsep,down%Rsep,.false.,param_check)
      else
        if(down%Dsep%n(3) .ne. down%Rsep%n(3)) then
          if(run%m_ith==0) write(0,("a,i4,a,i4,a")) &
     "When creating pcig gathers axis 3 (apxn) must be same for D(",&
    down%Dsep%n(3),") and R(",down%Rsep%n(3),")"
          param_check=.false.
        end if
        if(down%Dsep%n(4) .ne. down%Rsep%n(4)) then
          if(run%m_ith==0) write(0,("a,i4,a,i4,a")) &
    "When creating pcig gathers axis 4 (apyn) must be same for D(",&
    down%Dsep%n(4),") and R(",down%Rsep%n(4),")"
          param_check=.false.
        end if
      end if 
    end if 
     
       
      !SETUP THE REST OF THE STRUCTURES
      if(.not. wei_dfile) then 
        call init_sep3d(down%Dsep,"SCRATCH","COMPLEX","REGULAR",(/1,1,1,1,1/),ctag="D")
        call copy_axis(down%Wsep,1,down%Dsep,1)
        call copy_axis(down%Wsep,2,down%Dsep,2)
        call copy_axis(down%Wsep,3,down%Dsep,3)
        call copy_axis(down%Wsep,4,down%Dsep,4)
        call wei_param_axis(down%Dsep,"w_",5)
      end if
      if(.not. wei_rfile) then 
        call from_param("image_real",image_real,.false.)
        if(image_real) then
          call init_sep3d(down%Rsep,"SCRATCH","FLOAT","REGULAR",(/1,1,1,1,1/),ctag="R")
        else
          call init_sep3d(down%Rsep,"SCRATCH","COMPLEX","REGULAR",(/1,1,1,1,1/),ctag="R")
        end if
        call copy_axis(down%Wsep,1,down%Rsep,1)
        call copy_axis(down%Wsep,2,down%Rsep,2)
        if(.not. down%pCIG) then
          call copy_axis(down%Wsep,3,down%Rsep,3)
          call copy_axis(down%Wsep,4,down%Rsep,4)
        else
          call wei_param_axis(down%Rsep,"px",3)
          call wei_param_axis(down%Rsep,"py",4)
        end if
        call wei_param_axis(down%Rsep,"z_",5)
      end if
      down%Wsep%n(5:6)=(/down%nzs,down%nws/)
      only_zero_offset=.false.
      if(down%adj .and. run%operation(1:3)=="mig")&
         call from_param("only_zero_offset",only_zero_offset,.false.)

      !CHECK TO MAKE SURE THE SPACES MAKE SENSE
      if(.not. down%shot_profile) then
        call axis_check_param_file("amx",1,down%Wsep,down%Dsep,.false.,param_check)
        call axis_check_param_file("amy",2,down%Wsep,down%Dsep,.false.,param_check)
        call axis_check_param_file("ahx",3,down%Wsep,down%Dsep,.false.,param_check)
        call axis_check_param_file("ahy",4,down%Wsep,down%Dsep,.false.,param_check)
        call axis_check_param_file("amx",1,down%Wsep,down%Rsep,.false.,param_check)
        call axis_check_param_file("amy",2,down%Wsep,down%Rsep,.false.,param_check)
        if(.not. down%pCIG) then
          if(only_zero_offset) then
             down%Rsep%o(3)=0.; down%Rsep%n(3)=1
             down%Rsep%o(4)=0.; down%Rsep%n(4)=1
          else
            call axis_check_param_file("ahx",3,down%Wsep,down%Rsep,&
              .false.,param_check)
            call axis_check_param_file("ahy",4,down%Wsep,down%Rsep,&
              .false.,param_check)
          end if
        else
          if(down%Rsep%n(3) .ne. down%Wsep%n(3)) then
            if(run%m_ith==0) write(0,("a,i4,a,i4,a")) &
    "When creating pcig gathers axis 3 (apxn) must be same for R(",&
    down%Dsep%n(3),") and apx_n(",down%Wsep%n(3),")"
            param_check=.false.
          end if
          if(down%Rsep%n(4) .ne. down%Wsep%n(4)) then
            if(run%m_ith==0) write(0,("a,i4,a,i4,a")) &
      "When creating pcig gathers axis 4 (apyn) must be same for R("&
        ,down%Dsep%n(4),") and apy_n(",down%Wsep%n(4),")"
            param_check=.false.
          end if
        end if
      else
        call seperr("shot profile space checks unimplemented")
      end if
     
      call find_extent(down%Wsep,3,hext)
      if(0 < hext(1) .or. 0 > hext(2)) then
        if(run%m_ith==0) write(0,*) "offset axis (3) must contain 0"
        param_check=.false.
      end if
        
      call find_extent(down%Wsep,4,hext)
      if(0 < hext(1) .or. 0 > hext(2)) then
        if(run%m_ith==0) write(0,*) "offset axis (4) must contain 0"
        param_check=.false.
      end if
      call check_slow_compat(down%ssep,param_check)

      if(abs(down%rsep%o(5) -down%ssep%o(3))>epsilon(down%ssep%o(3))) then
        if(run%m_ith==0) write(0,*) "AxIS 5 of image o=",down%rsep%o(5)&
    ," and S =",down%ssep%o(3)," mismatch"
        param_check=.false.
      end if
      if(abs(down%rsep%d(5) -down%ssep%d(3))>epsilon(down%ssep%d(3))) then
        if(run%m_ith==0) write(0,*) "AXIS 5 of image d=",down%rsep%d(5),&
      " and axis 3 S =",down%ssep%d(3)," mismatch"
        param_check=.false.
      end if
      if(down%rsep%n(5)> down%ssep%n(3)) then
        if(run%m_ith==0) write(0,*) "AXIS 5 of image n=",down%rsep%n(5)&
      ," > than axis  3 S =",down%ssep%n(3)," mismatch"
        param_check=.false.
      end if
       
            
      if(down%conv_wave) then
         call check_slow_compat(down%s2sep,param_check)
        if(abs(down%rsep%o(5) -down%s2sep%o(3))>epsilon(down%s2sep%o(3))) then
          if(run%m_ith==0) write(0,*) "AXIS 5 of image o=",down%rsep%o(5)," and S =",down%s2sep%o(3)," mismatch"
          param_check=.false.
        end if
        if(abs(down%rsep%d(5) -down%s2sep%d(3))>epsilon(down%s2sep%d(3))) then
          if(run%m_ith==0) write(0,*) "AXIS 5 of image d=",down%rsep%d(5)," and axis 3 S =",down%s2sep%d(3)," mismatch"
          param_check=.false.
        end if
        if(down%rsep%n(5)> down%s2sep%o(n)) then
          if(run%m_ith==0) write(0,*) "AXIS 5 of image n=",down%rsep%o(5)," > than axis  3 S =",down%s2sep%o(3)," mismatch"
          param_check=.false.
        end if
      end if
   
      if(mod(down%dsep%n(5),down%wsep%n(6)) .ne. 0)  then
        if(run%m_ith==0) write(0,*) down%nws,"=nws  not multiple of Data-n(5)=",down%dsep%n(5)
        param_check=.false.
      end if
        
      if(mod(down%rsep%n(5),down%wsep%n(5)) .ne. 0)  then
        if(run%m_ith==0) write(0,*) down%nzs,"=nzs  not multiple of R-n(5)=",down%rsep%n(5)
        param_check=.false.
      end if

      if(down%wsep%n(1) .ne. 1) then
       if(product(down%wsep%n(2:3))==1 .and. down%wsep%n(4).ne. 1) then
          if(run%m_ith==0) write(0,*) "DATA n1,n4 >1 .and. n2,n3==1 not logical"
          param_check=.false.
      end if
     end if
     if(down%wsep%n(1)==1) then
       if(down%wsep%n(3).ne.1) then
          if(run%m_ith==0) write(0,*) "DATA n3 >1 .and. n1==1 not logical"
          param_check=.false.
       end if
     end if
    
          
        
  end function
subroutine copy_axis(sep1,iax1,sep2,iax2)
   type(sep3d) :: sep1,sep2
    integer :: iax1,iax2
    sep2%n(iax2)=sep1%n(iax1)
    sep2%o(iax2)=sep1%o(iax1)
    sep2%d(iax2)=sep1%d(iax1)
    sep2%label(iax2)=sep1%label(iax1)
    sep2%unit(iax2)=sep1%unit(iax1)
end subroutine
  subroutine check_slow_compat(struct,stat) 
   type(sep3d) ::  struct
   logical :: stat
   real :: mext(2),hext(2),sext(2),max_hext
      if(down%Wsep%n(1) > 1) then
        call find_extent(down%Wsep,1,mext)
        call find_extent(struct,1,sext)
        hext=0
        if(down%Wsep%n(3) > 1) call find_extent(down%Wsep,3,hext)
        max_hext=max(abs(hext(1)),abs(hext(2)))
        if(mext(1)-max_hext < sext(1)) then
           if(run%m_ith==0) write(0,*) trim(struct%tag),"=tag slowness: origin along axis 1 (",sext(1),") too large must be <= ",mext(1)-max_hext
           stat=.false.
        end if
        if(mext(2)+max_hext > sext(2)) then
           if(run%m_ith==0) write(0,*) trim(struct%tag),"=tag slowness: axis 1 too short (",sext(2),") must be >= ",mext(2)+max_hext
           stat=.false.
        end if
      end if 
      if(down%Wsep%n(2) > 1) then
        call find_extent(down%Wsep,2,mext)
        call find_extent(struct,2,sext)
        hext=0
        if(down%Wsep%n(4) > 1) call find_extent(down%Wsep,4,hext)
        max_hext=max(abs(hext(1)),abs(hext(2)))
        if(mext(1)-max_hext < sext(1)) then
           if(run%m_ith==0) write(0,*) trim(struct%tag),"=tag slowness: origin along axis 2 (",sext(1),") too large must be <= ",mext(1)-max_hext
           stat=.false.
        end if
        if(mext(2)+max_hext > sext(2)) then
           if(run%m_ith==0) write(0,*) trim(struct%tag),"=tag slowness: axis 2 too short (",sext(2),") must be >= ",mext(2)+max_hext
           stat=.false.
        end if
      end if 
            

  end subroutine

   subroutine axis_check_file(axisdes,iax,psep,fsep,exact,stat)
    type(sep3d) :: psep,fsep
    character(len=*) axisdes
    logical :: stat
    integer :: iax
    logical  :: exact
    real   :: mye
    mye=epsilon(mye)
    if(.not. exact) mye=mye*10.
    
    if(psep%n(iax) .ne. fsep%n(iax)) then
      if(run%m_ith==0) write(0,("a,i2,a,a,a,a,a,i3,a,a,a,i3,a")) "n of axis "&
      ,iax,"(",trim(axisdes),") of ",trim(fsep%tag),"(",fsep%n(iax),&
      ") and  ",trim(psep%tag),"(",psep%n(iax),") are inconsistent"
       stat=.false.
    end if 
    if(psep%n(iax) >1) then
      if(abs(psep%o(iax) - fsep%o(iax)) > mye) then
         if(run%m_ith==0) write(0,("a,i2,a,a,a,a,a,f12.5,a,a,a,f12.5,a")) &
   "o of axis ",iax,"(",trim(axisdes),") of ",trim(fsep%tag),"(",fsep%o(iax),&
        ") and  ",trim(psep%tag),"(",psep%o(iax),") are inconsistent"
         stat=.false.
       end if
      if(abs(psep%d(iax) - fsep%d(iax)) > mye) then
         if(run%m_ith==0) write(0,("a,i2,a,a,a,a,a,f12.5,a,a,a,f12.5,a")) &
     "do of axis",iax,"(",trim(axisdes),") of ",trim(fsep%tag),"(",fsep%d(iax),&
        ") and  ",trim(psep%tag),"(",psep%d(iax),") are inconsistent"
         stat=.false.
       end if
    end if 
   end subroutine

   subroutine axis_check_param_file(axisdes,iax,psep,fsep,exact,stat)
    type(sep3d) :: psep,fsep
    character(len=*) axisdes
    logical :: stat
    integer :: iax
    logical  :: exact
    real   :: mye
    mye=epsilon(mye)
    if(.not. exact) mye=mye*10.
    
    if(psep%n(iax) .ne. fsep%n(iax)) then
      if(run%m_ith==0) write(0,("a,i2,a,a,a,i3,a,a,a,i3,a")) "n of axis",iax," of ",trim(fsep%tag),"(",fsep%n(iax),&
      ") and ",trim(axisdes),"(",psep%n(iax),") are inconsistent"
       stat=.false.
    end if 
    if(psep%n(iax) >1) then
      if(abs(psep%o(iax) - fsep%o(iax)) > mye) then
         if(run%m_ith==0) write(0,("a,i2,a,a,a,f12.5,a,a,a,f12.5,a")) "o of axis",iax," of ",trim(fsep%tag),"(",fsep%o(iax),&
        ") and ",trim(axisdes),"(",psep%o(iax),") are inconsistent"
         stat=.false.
       end if
      if(abs(psep%d(iax) - fsep%d(iax)) > mye) then
         if(run%m_ith==0) write(0,("a,i2,a,a,a,f12.5,a,a,a,f12.5,a")) "d of axis",iax," of ",trim(fsep%tag),"(",fsep%d(iax),&
        ") and ",trim(axisdes),"(",psep%d(iax),") are inconsistent"
         stat=.false.
       end if
    end if 
   end subroutine



   subroutine find_extent(struct,iax,ext)
     type(sep3d) :: struct
     integer :: iax 
     real :: ext(2),beg,end

     beg=struct%o(iax)
     end=struct%o(iax)+(struct%n(iax)-1)*struct%d(iax)
     if(end > beg) then
       ext=(/beg,end/)
     else
       ext=(/end,beg/)
     end if
   end subroutine
   
  

  !----------------------------------------------------------------

  !READ IN PARAMETERS FROM PAR FILE
  logical function process_parameters()
   integer :: ierr

   if(isfile(down%VS))  then
     call init_sep3d(down%VS,down%Ssep,"SCRATCH")
   else
     call seperr("Must supply S (slowness/velocity file)")
   end if
   if(down%conv_wave) then
     call init_sep3d(down%VG,down%S2sep,"SCRATCH")
   end if
    if(wei_rfile) call init_sep3d("R",down%Rsep,"SCRATCH")
    if(wei_dfile) call init_sep3d("D",down%Dsep,"SCRATCH")
   call init_sep3d(down%Wsep,"SCRATCH","COMPLEX","REGULAR",(/1,1,1,1,1,1/))
   call wei_param_axis(down%Wsep,"mx",1)
   call wei_param_axis(down%Wsep,"my",2)
   call wei_param_axis(down%Wsep,"hx",3)
   call wei_param_axis(down%Wsep,"hy",4)
      process_parameters=T
  end function process_parameters
  !----------------------------------------------------------------
  !RESET DOWN VARIABLES
  subroutine down_reset(adj,add,mod_,dat_)
    logical,            intent(in) :: adj,add
    character(len=128)  :: mod_,dat_

    down%adj  = adj
    down%add  = add
    if(run%restart) down%add=.true.
    call pass_it(mod_)
    call pass_it(dat_)
!    write(0,*) "ASS RUULT",ruN%m_ith,trim(mod_),trim(dat_)
    down%mod_ = mod_
    down%dat_ = dat_

  end subroutine down_reset


  subroutine pass_it(name) 
    character(len=128) :: name,jj
    integer :: nsz,ierr
  
#ifdef SEP_MPI
    if(run%m_ith==0) nsz=len(trim(name))
    call MPI_Bcast(nsz,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
    jj=name
    call MPI_Bcast(jj,nsz,MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
    name=jj(1:nsz)
!    write(0,*) "CHECK2THIS",nsz,trim(name)
#endif

  end subroutine
  



  !----------------------------------------------------------------  
  !PRINT STATUS 
  subroutine down_report()
    call weiline()
!    write(run%io_unit,*)                    "        *       model=",down%mod_
!    write(run%io_unit,*)                    "        *        data=",down%dat_
    call weiline()
    write(run%io_unit,*)                    "        *        pCIG=",down%pCIG
    write(run%io_unit,'(a,i5)')             "         *         nzs=",down%nzs
    write(run%io_unit,'(a,i5)')             "         *         nws=",down%nws
    write(run%io_unit,'(a,i5)')             "         *         nzb=",down%nzb
    write(run%io_unit,'(a,i5)')             "         *         nwb=",down%nwb
    write(run%io_unit,'(a,i5)')             "         *         nref=",down%nref
    call weiline()
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Wmx= ",down%wsep%n(1),  down%wsep%o(1),  down%wsep%d(1)
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Wmy= ",down%wsep%n(2),  down%wsep%o(2),  down%wsep%d(2)
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Whx= ",down%wsep%n(3),  down%wsep%o(3),  down%wsep%d(3)
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Why= ",down%wsep%n(4),  down%wsep%o(4),  down%wsep%d(4)
    if(down%pciG) then
      write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Rpx= ",down%rsep%n(3),  down%rsep%o(3),  down%rsep%d(3)
      write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Rpy= ",down%rsep%n(4),  down%rsep%o(4),  down%rsep%d(4)
    end if
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Rz= ",down%rsep%n(5),  down%rsep%o(5),  down%rsep%d(5)
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *         Dw= ",down%dsep%n(5),  down%dsep%o(5),  down%dsep%d(5)
    call weiline()
  end subroutine down_report

  !SET THE NUMBER OF BLOCKS WE WILL BE USING
  subroutine set_blocks(n_wb,n_zb,n_blocks)
   integer,optional :: n_zb,n_wb,n_blocks
   if(present(n_zb)) nzb=n_zb
   if(present(n_wb)) nwb=n_wb
   if(present(n_blocks)) nblocks=n_blocks
  end subroutine


  !----------------------------------------------------------------
  !PRINT HOW MUCH WE HAVE FINISHED
  subroutine wei_followup(izb,iwb)
    integer  :: pct
    integer :: iwb,izb
     pct=real(idone)/real(nblocks)*100.
     if(pct>nprint) then
       write(run%io_unit,'( "         *&
        & done ",f5.2,"%","    izb=",i4,"/",i4,"     iwb=",i4,"/",i4," thread=",i4)') &
        real(idone)/real(nblocks)*100.,izb,nzb,iwb,nwb,run%m_ith
        nprint=nprint+dprint
     end if
     idone=idone+1
  end subroutine wei_followup
  !----------------------------------------------------------------
  !CONVERT TO RADIANS
  subroutine process_radians()
   
    down%freqs=2*pi*down%freqs


  end subroutine process_radians

  subroutine init_progress_count()
    idone=0
  end subroutine
  !MAKE SURE THAT THE NUMBER OF OFFSETS AND P ARE THE SAME
  !----------------------------------------------------------------
end module wei_down
