! Information on the  job
! This includes  what job is being run
! Thread number OMP/MPI,etc
module wei_run
  use wei_external
  use wei_util
  use wei_timers_mod
  use weirun_type_mod
  use wei_mem_mod
#ifdef SEP_MPI
  use mpi_sep 
#endif
use sep
  implicit none
  complex , pointer  :: small_buf(:,:)
  logical,save :: wei_dfile,wei_rfile

  contains
!----------------------------------------------------------------
!
!
! Function for initializing the run enviro
!
!

  logical function weirun_init()
    character(len=128) :: operator,operation
    integer :: nsize,ierr
    integer, external :: getch

#ifdef DEBUG
    character(len=128), parameter :: unit='weirun_init'
    if(run%debug) call in(unit)
#endif DEBUG
  
    if(.not. init_mem_usage()) then
       write(0,*) "trouble initializing memory usage"
       return
    end if

    if(.not. init_wei_timers()) then
      write(0,*) "trouble initializing timers"
      return
    end if

    !Initialize parallel environment for run
    if(.not. run_parallel()) then
       write(0,*) "trouble initializing the wei_run_parallel"
       return
    end if

    call from_param('operation',operation,'migration...')
    run%operation = operation

    if(.not. setup_log_datapath()) then
        write(0,*) "trouble setting up output/log file"
        return
     end if
    
    call from_param('operator',operator,'.....')
    run%operator = operator

    call from_param('verb', run%verb_all, F)
    call from_param('verb', run%verb, F)
    call from_param('debug',run%debug,F)
    call from_param('restart',run%restart,F)
    call from_param('depar',run%depar,F)
    call from_param("unlink",run%unlink,T)
    if(      run%debug) run%verb =T
    if(.not. run%debug) run%depar=T

    !! whether or not to save memory
    call from_param("mem_save",run%mem_save,0)
    call from_param("copy_buf",nsize,5);nsize=nsize*1000*1000/8
    allocate(small_buf(nsize,2))
     

    weirun_init=F

    if(run%mem_save > 0 .and. run%operation(1:3)/="mig" ) then
      write(0,*) "memory save option only available for migration"
      return
    end if 
      

   call from_param("io_file",run%file_out,"NONE")
   if(run%file_out(1:4)=="NONE") then
     run%io_unit=0
   else 
     run%io_unit=6
     open(run%io_unit,file=run%file_out)
   end if

   weirun_init=T

   if(run%m_ith.ne. 0) run%debug=.false.
   if(run%m_ith.ne. 0) run%verb=.false.
   if(run%verb) call run_report()

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function weirun_init


    

!----------------------------------------------------------------
!
!
! Subroutine to initialize OMP run environment
!
! Sets number of omp threads
!
!
!
!

  logical function run_omp_parallel()
#ifdef DEBUG
    character(len=128), parameter :: unit='omp_run_parallel'
    if(run%debug) call in(unit)
#endif

#ifdef SEP_OMP
    !$OMP PARALLEL
write(0,*) "IN OMP 2",omp_get_num_threads();
    run%o_nth = omp_get_num_threads()
    !$OMP END PARALLEL

    run%omp=T
#else
    run%o_nth=1
    run%omp=F
#endif

   run_omp_parallel=.true.



#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function  run_omp_parallel

!----------------------------------------------------------------
!
! Function for setting up mpi parallel enviro for mpi
!
! Turns of verb, debug except for the master thread
! Sets up the local datapath
! Figues out the extra to add for each thread
! records the number of threads and this threads num
!
 logical function runmpi_parallel()
   character(len=128) :: temp_ch,machine
   integer :: ierr,ilen,igrab,i
#ifdef DEBUG
    character(len=128), parameter :: unit='runmpi_parallel'
    if(run%debug) call in(unit)
#endif

    runmpi_parallel=F

#ifdef SEP_MPI
    call MPI_Get_processor_name(machine,ilen,ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, run%m_nth,ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, run%m_ith,ierr)
    call from_param("share_image",run%share_image,.false.)
#else
    run%share_image=.false.
    run%m_nth=1
    run%m_ith=0
#endif

    !TURN OFF VERB AND DEBUG FOR ALL BUT ZERO MPI THREAD
    if(run%m_ith /= 0) then
       run%verb=F
       run%debug=F
    end if


#ifdef SEP_MPI
    if(run%m_nth>1) then
      do i=1,len(run%local_path)-1
        run%local_path(ilen:ilen)=" "
      end do
       igrab=getch("local_datapath","s",run%local_path)

!       if(0==igrab) run%local_path="/net/"//machine(1:ilen)//"/scr1/temp/"
       if(0==igrab) run%local_path="/scr1/temp/"

       run%mach=trim(machine(1:8))
    ilen=len(trim(run%local_path));
    if(igrab==1) run%local_path(ilen:ilen)=" "


       if(run%m_ith<10) then
          write(run%tag_num,"(a,i1)") "_",run%m_ith
       else if(run%m_ith <100) then
          write(run%tag_num,"(a,i2)") "_",run%m_ith
       else
          write(run%tag_num,"(a,i3)") "_",run%m_ith
       end if
    end if
    run%mpi=T
    if(run%m_nth==1) run%mpi=F
#else
    run%mpi=F
#endif


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif

  runmpi_parallel=T

  end function runmpi_parallel


 subroutine run_report()
    write(run%io_unit,*) "operator=",trim(run%operator(1:4))," operation=",trim(run%operation(1:8))
    write(run%io_unit,*) "writing to device ",run%io_unit
    write(run%io_unit,*) "verbose=",run%verb," debug=",run%debug," para debug", run%depar
    write(run%io_unit,*) "restart=",run%restart
    write(run%io_unit,*)  "mpi",run%mpi," mpi_ithread=",run%m_ith," mpi_nthread=",run%m_nth
#ifdef SEP_OMP
    write(run%io_unit,*)  "omp",run%omp," omp_nthread=",run%o_nth
#else
    write(run%io_unit,*)  "not running with omp"
#endif
 end subroutine


!----------------------------------------------------------------
!
!  Subroutine for initializing parralel run (omp and mpi)
!
!
!
!
  logical function run_parallel()
     logical :: dude
#ifdef DEBUG
    character(len=128), parameter :: unit='run_parallel'
    if(run%debug) call in(unit)
#endif
 
    run_parallel=F

    if(.not. run_omp_parallel()) then
      write(0,*) "trouble initializing omp enviro"
      return
    end if

     dude=runmpi_parallel()
    if(.not. dude) then
      write(0,*) "trouble initializing mpi enviro"
      return
    end if
    
    run_parallel=T
#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function run_parallel

logical function setup_log_datapath()
 integer :: id,ir
 real    :: a 
 integer, external :: getch
 integer :: ierr
 logical :: new_datapath
  character(len=1024) :: temp_ch,temp2_ch
#ifdef DEBUG
    character(len=128), parameter :: unit='setup_log_output'
    if(run%debug) call in(unit)
#endif DEBUG

   setup_log_datapath=.false.
 
  
    run%logfile="logfile"
    if(run%operation(1:3)=="mig" .or. run%operation(1:3)=="inv") run%logfile="R"
    if(run%operation(1:3)=="mod") run%logfile="D"
    call from_param("logfile",run%logfile,run%logfile)
     id=0
     ir=0

#ifdef SEP_MPI
if(run%m_ith==0) then
#endif
     if(1==getch("output_datapath","s",temp2_ch)) then
       new_datapath=.true.
       temp_ch="datapath="//trim(temp2_ch)
       call getch_add_string(temp_ch)
     else 
       new_datapath=.false.
     end if

     if(isfile("D"))  then
       call auxinout("D")
       id=1
     end if
     if(isfile("R"))  then
       call auxinout("R")
       ir=1
     end if
     call set_yes_putch()
     call set_alternate_putch(run%logfile)
     if((trim(run%logfile)=="D" .and.  id==0) .or. &
       (trim(run%logfile)=="R" .and.  ir==0) ) then
       call auxout(trim(run%logfile))
       ierr=srite(trim(run%logfile),a,4)
!       if(new_datapath) ierr=srite(trim(run%logfile),a,4)
       call auxclose(trim(run%logfile))
       call auxinout(trim(run%logfile))
     end if
     call to_history("operation",run%operation,run%logfile)
#ifdef SEP_MPI
end if
  call MPI_BCAST(ir,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(id,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
#endif
  wei_rfile=(ir==1)
  wei_dfile=(id==1)
  if(run%m_ith /=0 ) call auxinout("R")
  if(run%m_ith /=0 ) call auxinout("D")


 write(temp_ch,"(a,a,a,a)")"datapath=",trim(run%local_path),trim(run%tag_num),&
       "DATA_"
 call getch_add_string(temp_ch)

   setup_log_datapath=.true.

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
end function


logical function make_local_tag(tag_out,tag_in,add)
  character(len=*) :: tag_in,tag_out,add
  character(len=256) :: temp
  integer :: i,j 

  
    do i=1,64
      temp(i:I)=" "
    end do

    if(tag_in(1:1)/="/") then
      write(tag_out,*) trim(run%local_path),trim(tag_in),trim(add)
    else
      write(tag_out,*) trim(tag_in),trim(add)
    end if

   call clean_string(tag_out)
 
   make_local_tag=.true.
end function



end module wei_run
