module wei_cam
  !! Wave-Equation Migration
  !! Common-azimuth

  use wei_util
  use wei_external
  use wei_down
  use wei_run
  use wei_obj_S
  use wei_kmap
  use wei_fft
  use wei_fk_share_mod

  !---------------
  implicit none

  real,private                            :: dkxko, cdkxko,idkzko,rdkzko
  integer     ,private                    :: nkxko
  real                            ::  kxko
  real,    dimension(:), pointer,private :: ckzko,exp_kzko,ikzko
  complex, dimension(:), pointer,private ::  kzko, tkzko
  complex, pointer, private :: w_extrap_1(:,:,:,:,:),w_extrap_2(:,:,:,:,:)
  real, dimension(:) , pointer,private :: khy(:,:)
contains
!----------------------------------------------------------------
!!$=head1 NAME
!!$
!!$weicam_init -initialize  common azimuth migration
!!$
!!$=head1 SYNOPSIS
!!$
!!$ierr= weicam_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$Common Azimuth Migration Initialization operator
!!$
!!$Grabs kxko=2. and nkxko=2001 from par
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weicamo> L<weiwem_init>
!!$
!!$=cut
!!$
  subroutine comaz_init(max,n)
    integer, intent(in) :: n
    real,    intent(in) :: max
    integer             :: er
    integer             :: i,m
    real                :: kxko

     allocate(khy(down%wsep%n(1),down%wsep%n(3)))
    allocate(ckzko(n),stat=er)
    call add_mem(size(ckzko))
    if(er/=0) call seperr('cant allocate ckzko')


    cdkxko = max/(n-1)
    m=n/max

    do i=1,m
       kxko     = 1. * (i-1)/(m-1) !! kx/ko=0...1
       ckzko(i) = sqrt(1-kxko**2)
    end do
    do i=m+1,n
       ckzko(i) = 0
    end do

!    ckzko(n) = ckzko(n-1) !! avoid kzs=0 or kzr=0
    ckzko = .5 * ckzko
  end subroutine comaz_init
  !----------------------------------------------------------------



  integer function weicam_gsp_init() result(st)
    integer :: er,n
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam_init'
    if(run%debug) call in(unit)
#endif

     n=size(tmp_on,5)
     if(run%fk_extra .ne. 2) call seperr("expecting  fk_extra ==2 for gsp")
     w_extrap_1=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
     n=n-run%o_nth
     w_extrap_2=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
     OK=weicam_init()
     down%maxslow=.true.

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weicam_gsp_init



!!$=head1 NAME
  integer function weicam_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam_init'
    if(run%debug) call in(unit)
#endif

    !! init SSR
    call from_param("nkxko",nkxko,2001)
    call from_param( "kxko", kxko,2.)
    call dsr_init(kxko,nkxko,kzko,tkzko,exp_kzko,ikzko,dkxko,nkxko,idkzko,rdkzko)
    call  comaz_init(kxko,nkxko)

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weicam_init



!!$=head1 NAME
!!$
!!$weicam - one depth step of common azimuth migration
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weicam(iws,izs,ifk,ith,wfld)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item iws - integer
!!$
!!$  current frequency slice
!!$
!!$=item izs - integer
!!$
!!$  current depth slice
!!$
!!$=item ifk - integer
!!$
!!$  current fk operation
!!$
!!$=item ith - integer
!!$
!!$  current thread
!!$
!!$=item wfld - complex(:,:,:,:,:)
!!$
!!$  wavefield
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$Common Azimuth Migration  operator
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<wei_caminit> L<weiwem>
!!$
!!$=cut
  !----------------------------------------------------------------
  integer function weicam_gsp(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) ::iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    :: w
    real    :: inv_norm_ko,inv_norm_ko2
    real    :: r_inv_norm_ko,i_inv_norm_ko
    integer :: imx,imy,ihx,ihy 
    integer :: is,ir,i,sign
    complex :: ikz
    real    :: kzs,kzr,k_y
    real    :: ksy,kry,k,j,kmy
    real    :: kss,krr,ko,fract,iko2,cract,inv_slow_cb
    real,pointer :: kx(:,:),ky(:,:)
    !NEW
    real :: slow_0,epsi,slow_0_sq,w_data,kow,k_sq_comaz,dz
    real :: a0,dz_w,d,myr,myi
    real :: inv_slow_0,kow_sq,inv_gamma
    real :: det,gamma,d3
    real  :: smin,smax,bell_sq,kowth,kowth2
!BEGIN ORDER 2
    real :: inv_slow_0_cb
    complex :: c_etak2,c_shift2,rb,c_gamma_cb,cinv_gamma_cb,inv_gamma_sq,inv_gamma_cb
!END ORDER2
    complex :: cphase,c_gamma,c_det,c_kow,c_kow_sq,inv_gmma
    complex :: c_shift,cinv_gamma,c_etak1
    complex :: ra,c_shift1,s
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam'
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%fk_t)
#endif

!! set constants




    w= down%w_local(iws)
   call calc_comp_depend_vars(w,slow_0,ko,kx,ky,sign,slo%th(ith)%source,izs,ifk)

    epsi=epsilon(1.)
    inv_norm_ko=1./(ko*cdkxko);
    inv_norm_ko2=1./(ko*ko)
    iko2=1./ko**2
    r_inv_norm_ko=1./(rdkzko)
    i_inv_norm_ko=1./(idkzko)


  slow_0_sq=slow_0*slow_0
  inv_slow_0=1.0/slow_0
  inv_slow_0_cb=1.0/slow_0**3
  w_data=w
  dz_w=var%dzb(izs)*w_data

  call weiift(ith,izs,ith,w_extrap_1)
!  call weiift(ith,izs,ith,w_extrap_2)


  a0=-dz_w
  s=cmplx(0,-w)
  dz=var%dzb(izs)

  w_extrap_1(:,:,:,:,ith)=s/2.*dz*w_extrap_1(:,:,:,:,ith)
!  w_extrap_2(:,:,:,:,ith)=-s/8.*dz*w_extrap_2(:,:,:,:,ith)
   



  cract=slow_0
  call calc_sign_depend_vars(fract,tkzko,var%dzb(izs),kzko,ko)
  call calc_khy(ko,inv_norm_ko,izs)
   

  kowth=sin(5.*pi/12.0)  !60 degrees
  kowth2=sin(5.2*pi/12.0)  !75 degrees



  !kow =sqrt(k_sq_comaz)/w
  ! kow *w = sqrt(k_sq_comaz)
  ! kow*kow*w*w = sqrt(k_sqrt_comaz)
!  kowth=kowth*kowth*w*w
!  kowth2=kowth2*kowth2*w*w
  d3=sqrt(kowth)-sqrt(kowth2)

if(ko >0.) then 
  do ihx=1,size(wfld,3)
    do imy=1,down%wsep%n(2) 
      kmy=kmap%kmy(imy)
      do imx=1,down%wsep%n(1) 
        if (abs(wfld(imx,imy,ihx,1,iws))>0.0) then
              ! evaluate DSR with common-azimuth k_hy
              k_y=kmy+sign*kmy*khy(imx,ihx)
              k_sq_comaz=kx(imx,ihx)**2+k_y**2
!              kow=sqrt(k_sq_comaz)/w
!              if ((kow>kowth) .and. (kow<kowth2)) then
!                 bell_sq=cos(kow-kowth/d3*pi/2.0)
!                 bell_sq=bell_sq*bell_sq
!!                 bell_sq=0.0
!              else if (kow<kowth) then
!                 bell_sq=1.0
!              else
!                 bell_sq=0.0
!              end if
 
                 k = iko2*k_sq_comaz
                 if (k<1.) then
                    i= min(nkxko,max(1, int(1 + k*i_inv_norm_ko)))
                    gamma=(ikzko(i)*cract )
                    inv_gamma=1./max(gamma,.000001)
                    j=fract*ikzko(i)
                    c_shift=cmplx(cos(j),sin(j))
                    c_etak1=inv_gamma-inv_slow_0
                    c_etak2=(inv_gamma**3-inv_slow_0_cb)
                 else
                    i=max(1, min(nkxko,int(1 + (k-1.)*r_inv_norm_ko)))
                    inv_gamma=1.0/max(.000001,(cract*exp_kzko(i)))
                    j=fract*exp_kzko(i) 
                    c_shift=cmplx(exp(j), 0.0)
                    c_etak1=cmplx(inv_gamma, -inv_slow_0)
                    c_etak2=cmplx(inv_gamma**3,-inv_slow_0_cb)
                 end if
                    
                   

!!BEGIN ORDER 1
                 ra = (w_extrap_1(imx,imy,ihx,1,ith)*c_etak1) &
                      / wfld(imx,imy,ihx,1,iws)
!END ORDER 1
!BEGIN ORDER 2
!                ra = ((w_extrap_1(imx,imy,ihx,1,ith)*c_etak1) &
!                      + (w_extrap_2(imx,imy,ihx,1,ith)*c_etak2)) &
!                      / wfld(imx,imy,ihx,1,iws)
!END ORDER2
!                 c_shift1=cexp(cmplx(0.,aimag(ra)))
!                 rb=1+real(ra)/cmplx(1.0,aimag(ra))
!                 if(abs(rb)>0.) then
!                     c_shift2=rb/abs(rb)
!                 else
!                     c_shift2=rb
!                 end if
                  
                 myr=real(ra)
                 myi=aimag(ra)
                 ra=1+myr/cmplx(1,myi)
                 c_shift1=cexp(cmplx(0.0,myi))*ra/cabs(ra)
                 wfld(imx,imy,ihx,1,iws) = &
                 wfld(imx,imy,ihx,1,iws)*c_shift*c_shift1!*c_shift2



              else
                 wfld(imx,imy,ihx,1,iws) = cmplx(0.0,0.0)
              end if


              ! GSP stuff finishes here
           end do 
        end do 
     end do 
  end if 


!  if(run%m_ith==0) write(0,*) "CHECK THIS",sum(cabs(wfld(:,:,:,:,iws)))
!
#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%fk_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weicam_gsp
  !----------------------------------------------------------------
  integer function weicam(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) ::iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    :: w
    real    :: inv_norm_ko,inv_norm_ko2
    real    :: r_inv_norm_ko,i_inv_norm_ko
    integer :: imx,imy,ihx,ihy 
    integer :: is,ir,i,sign
    complex :: ikz
    real    :: kzs,kzr,k_y
    real    :: ksy,kry,k,j,kmy
    real    :: kss,krr,ko,fract,iko2,slow_0
    real,pointer :: kx(:,:),ky(:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='weicam'
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%fk_t)
#endif

!	if(run%m_ith==0) write(0,*)"in weicam",sum(cabs(wfld(:,:,:,:,ith)))

    w= down%w_local(iws)

    !FIGURE OUT WHAT SOURCE WE ARE USING
    call calc_comp_depend_vars(w,slow_0,ko,kx,ky,sign,slo%th(ith)%source,izs,ifk)
   

    inv_norm_ko=1./(ko*cdkxko)
    iko2=1./ko**2
    inv_norm_ko2=1./(ko*ko)
    r_inv_norm_ko=1./(rdkzko)
    i_inv_norm_ko=1./(idkzko)
    call calc_sign_depend_vars(fract,tkzko,var%dzb(izs),kzko,ko)



  call calc_khy(ko,inv_norm_ko,izs)
!khy=1.
   
if(ko >0.) then
    do ihy=1,size(wfld,4)
       do ihx=1,size(wfld,3)
          do imy=1,down%wsep%n(2)
             kmy=kmap%kmy(imy)
             do imx=1,down%wsep%n(1)
                k_y =kmy+ sign*khy(imx,ihx)*kmy
!                k_y=0
                k = iko2*(kx(imx,ihx)**2 + k_y**2)
                i =max(1, min(int(1 + k/ko / dkxko) , nkxko))
                ikz= ko*tkzko(i)

                   wfld(imx,imy,ihx,ihy,iws) = &
                   wfld(imx,imy,ihx,ihy,iws) * cexp(var%dzb(izs)*ikz)
               
!                if(k < 1.) then
!                    i= min(nkxko,max(1, int(1 + k*i_inv_norm_ko)))
!                    j=fract*ikzko(i)
!                   wfld(imx,imy,ihx,ihy,iws) = wfld(imx,imy,ihx,ihy,iws) * &
!                     cmplx(cos(j),sin(j))
!                else
!                   i=max(1, min(nkxko,int(1 + (k-1.)*r_inv_norm_ko)))
!                   j=fract*exp_kzko(i)
!                   wfld(imx,imy,ihx,ihy,iws) = &
!                   wfld(imx,imy,ihx,ihy,iws) * exp(j)
!                end if

             end do
          end do
       end do
    end do
end if

!	if(run%m_ith==0) write(0,*)"cam result",sum(cabs(wfld(:,:,:,:,ith)))

#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%fk_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weicam
  !----------------------------------------------------------------

subroutine calc_khy(ko,inv_norm_ko,izs)
integer :: is,ir,ihx,imx,izs
real :: kzs,kzr,ko,inv_norm_ko
 do ihx=1,down%wsep%n(3)
   do imx=1,down%wsep%n(1)
        is = min(int(1 + abs(kmap%ksx(imx,ihx)*inv_norm_ko)) , nkxko)
        ir = min(int(1 + abs(kmap%krx(imx,ihx)*inv_norm_ko)) , nkxko)
        kzs = ko*ckzko(is)
        kzr = ko*ckzko(ir)
        if((kzs > 0.) .and. (kzr >0.))then 
           khy(imx,ihx) =(kzr-kzs)/(kzs+kzr)
        else
           khy(imx,ihx) = 0.
       end if
   end do
 end do
end subroutine



end module wei_cam
