module wei_image
  !! WEI imaging condition

  use wei_util
  use wei_down
	use wei_cig
  !---------------
  implicit none

  complex, dimension(:,:,:,:,:), pointer, private :: imwf
	integer, private :: begpadx,begpady
contains
  !----------------------------------------------------------------  
  subroutine weiimage_init()
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weiimage_init'
    if(run%debug) call in(unit)
#endif

		logical :: pcig
    integer :: npadx,npady

    call from_param("pCIG",pcig,.false.)
    call from_param("pcigx_pad",npadx,0)
    call from_param("pcigy_pad",npady,0)

	
!if(run%m_ith==0) write(0,*) "CEHCK1",npadx,npady

    begpadx=npadx/2
    begpady=npady/2

		if(npadx > 0 .or. npady > 0) then
			allocate(imwf(down%wsep%n(1),down%wsep%n(2),down%wsep%n(3)+npadx,down%wsep%n(4)+npady,run%o_nth),stat=er)
      if(er/=0) then
				 call seperr('cant allocate imwf')
    	end if
		else
    	!! imaging wavefield
    	if(run%mem_save<2) then
     	 imwf=>tmp_on(:,:,:,:,1:run%o_nth)
    	end if
		end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weiimage_init
  !----------------------------------------------------------------  
  subroutine weiimgW2R(izs,iws,ith,image,wfld,IGop)
    integer, intent(in) :: izs,iws,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld, image
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weiimgW2R'
#endif
    interface !!-------------------------
       integer function IGop(wfld,iws,ith) result(st)
         complex, dimension(:,:,:,:,:), pointer :: wfld
         integer, intent(in) :: iws,ith
       end function IGop
    end interface !!---------------------
    integer :: i1,i2,i3,i4
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
            call start_timer_num(timer%image_t)
#endif


    if(run%mem_save>1) then
      st = IGop(wfld,iws,ith) !! CIG h->h or h->ph
      image=image/down%dsep%n(5)
    else
			imwf=0
      imwf(:,:,1+begpadx:begpadx+size(wfld,3), &
							 1+begpady:begpady+size(wfld,4),ith) = wfld(:,:,:,:,iws)
      st = IGop(imwf,iws,ith) !! CIG h->h or h->ph
!if(run%m_ith==0) write(0,*) "CEHCK",size(wfld,3),size(image,3),begpadx

      do i4=1,size(image,4)
      do i3=1,size(image,3)
      do i2=1,size(image,2)
      do i1=1,size(image,1)
      !$OMP ATOMIC
      image(i1,i2,i3,i4,izs) = &
      image(i1,i2,i3,i4,izs) + &
      imwf (i1,i2,i3+begpadx,i4+begpady,ith)/down%dsep%n(5)
     end do
     end do
     end do
     end do
    end if


#ifdef SEP_TIMER
            call stop_timer_num(timer%image_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weiimgW2R
  !----------------------------------------------------------------
  subroutine weiimgR2W(izs,iws,ith,image,wfld,IGop)
    integer, intent(in) :: izs,iws,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld, image
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weiimgR2W'
#endif
    interface !!-------------------------
       integer function IGop(wfld,iws,ith) result(st)
         complex, dimension(:,:,:,:,:), pointer :: wfld
         integer, intent(in) :: iws,ith
       end function IGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
            call start_timer_num(timer%image_t)
#endif

		imwf=0
    imwf(:,:,1+begpadx:size(image,3)+begpadx, &
						 begpady+1:begpady+size(image,4),ith) = image(:,:,:,:,izs)
    st = IGop(imwf,iws,ith) !! CIG: h->h or ph->h


    wfld(:,:,:,:,iws) = &
    wfld(:,:,:,:,iws) + &
    imwf(:,:,1+begpadx:begpadx+size(image,3),1+begpady:begpady+size(image,4),ith)/down%dsep%n(5)


#ifdef SEP_TIMER
            call stop_timer_num(timer%image_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weiimgR2W
  !----------------------------------------------------------------
end module wei_image





