module wei_kxmig
  !! fk-fx migration
  !! receive wfld at (izs,iws,ith)
  !! return  wfld downward/upward continued
  !! needs FKop, FXop

  use wei_util
  use wei_run
  use wei_external
  use wei_down
  use wei_obj_S
  use wei_fft
  use wei_kmap
  !---------------
  implicit none

  !! temporary interpolation arrays
  complex, dimension(:,:,:,:,:), pointer, private :: tmp1,tmp2
  complex, dimension(:,:), pointer, private :: tmps
  character(len=128) :: twave


  interface weikxmig
    module procedure weikxmig_1wave,weikxmig_2wave
  end interface



contains



  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine weikxmig_1wave(wfld,iws,izs,ith,KXMIGop,FKop,FXop)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer,external,optional ::    FKop,FXop
    integer :: izs,iws,ith
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmig'
#endif
    interface !!-------------------------
       integer function KXMIGop(wfld,iws,izs,ith,FKop,FXop) result(st)
         complex, dimension(:,:,:,:,:), pointer :: wfld
         integer, intent(in)       :: iws,izs,ith
         integer,external,optional :: FKop,FXop
       end function KXMIGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    st = KXMIGop(wfld,iws,izs,ith,FKop,FXop)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weikxmig_1wave




  !----------------------------------------------------------------
  subroutine weikxmig_2wave(wfld_r,wfld_s,iws,izs,ith,KXMIGop,FKop,FXop)
    complex, dimension(:,:,:,:,:), pointer         :: wfld_s,wfld_r
    integer,external,optional ::    FKop,FXop
    integer :: izs,iws,ith
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmig'
#endif
    interface !!-------------------------
       integer function KXMIGop(wfld_r,wfld_s,iws,izs,ith,FKop,FXop) result(st)
         complex, dimension(:,:,:,:,:), pointer ::wfld_s,wfld_r
         integer, intent(in)       :: iws,izs,ith
         integer,external,optional :: FKop,FXop
       end function KXMIGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    st = KXMIGop(wfld_r,wfld_s,iws,izs,ith,FKop,FXop)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine 

 !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_shot(wfld_rec,wfld_src,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld_src,wfld_rec
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

      !D.C. SOURCE WAVEFIELD
      call sellect_slow_comp(.true.,ith)!!selects slowness from source field
      st=weikxmigN_single(wfld_src,iws,izs,ith,Fkop,Fxop)
      !D.C. RECEIVER WAVEFIELD
      call sellect_slow_comp(.false.,ith)!!selects slowness from recvr field
      st=weikxmigN_single(wfld_rec,iws,izs,ith,Fkop,Fxop)

    end function weikxmigN_shot
  !----------------------------------------------------------------

  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_combo(wfld,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

  !LOOP OVER SOURCES
    !SELECT THE SOURCE
    call sellect_slow_comp(.true.,ith)
      st=weikxmigN_single(wfld,iws,izs,ith,Fkop,Fxop)
  end function weikxmigN_combo
!----------------------------------------------------------------







  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_mid(wfld,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

  !LOOP OVER SOURCES
  do isource=1,2 !{
    !SELECT THE SOURCE
    call sellect_slow_comp(isource==1,ith)
      st=weikxmigN_single(wfld,iws,izs,ith,Fkop,Fxop)
  end do
  end function weikxmigN_mid
!----------------------------------------------------------------










  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN_init'
    if(run%debug) call in(unit)
#endif

      !INITIALIZE TEMPORARY SPACE FOR WAVEFIELDS 
      call init_wave_temps()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weikxmigN_init


  !----------------------------------------------------------------
  !! 1 reference slowness
  integer function weikxmigN_single(wfld,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
            if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%kxmig_t)
#endif

    if(down%adj) then !! MIGRATION {
       !----------------------------------------------------------------       

       !REDO THE KMAPS IN CASE WE HAVE CHANGED THE OFFSET RANGE
       !LOOP OVER SOURCES
          !SELECT THE SOURCE
          call store_wavefield(wfld,ith,iws,.true.) !STORE THE CURRENT WAVEFIELD
          do ifk=1,slo%th(ith)%nrs(izs) !{
             call grab_part_wvfld(wfld,ifk,ith,iws,izs) !GRAB A PORTION OF THE WFLD
             st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
             call weiift(iws,izs,    ith,wfld)
             st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
             call weifft(iws,izs,    ith,wfld)
             call add_wavefield(wfld,ith,iws)
          end do !}
          call grab_wavefield(wfld,iws,ith,.true.) !GRAB THE ENTIRE WAVEFIELD

       !----------------------------------------------------------------
       !}
    else             !! MODELLING {
       !----------------------------------------------------------------

       !REDO THE KMAPS IN CASE WE HAVE CHANGED THE OFFSET RANGE

          call store_wavefield(wfld,ith,iws,.false.)


          do ifk=1,slo%th(ith)%nrs(izs) !{
             call grab_wavefield(wfld,iws,ith,.true.)
             call weiift(iws,izs,    ith,wfld)
             st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
             call weifft(iws,izs,    ith,wfld)
             st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op

             call store_part_wavefield(wfld,ifk,ith,iws,izs) 
          end do !}
          call grab_wavefield(wfld,iws,ith,.false.);
          !----------------------------------------------------------------
    end if!}

#ifdef SEP_TIMER
            if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%kxmig_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weikxmigN_single






  !----------------------------------------------------------------  




!---------------------------------------------
!
!
!  Initialize storage for wavefields
  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmig1(wfld,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
            if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%kxmig_t)
#endif

    !REDO THE KMAPS IN CASE WE HAVE CHANGED THE OFFSET RANGE
    if(down%adj) then !! MIGRATION {
       !----------------------------------------------------------------       
         call sellect_slow_comp(.true.,ith)
!         st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
!         call sellect_slow_comp(.false.,ith)
!         st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
!         call weiift(iws,izs,    ith,wfld)
!         st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
!         call sellect_slow_comp(.true.,ith)
!         st   = FKop(iws,izs,ifk,ith,wfld) !! f-x op
!         call weifft(iws,izs,    ith,wfld)
       !----------------------------------------------------------------
       !}
    else             !! MODELLING {
       !----------------------------------------------------------------
!         call weiift(iws,izs,    ith,wfld)
!         call sellect_slow_comp(.true.,ith)
!         st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
!         call sellect_slow_comp(.false.,ith)
!         st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
!         call weifft(iws,izs,    ith,wfld)
!         st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
!         call sellect_slow_comp(.true.,ith)
!         st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
    end if!}

#ifdef SEP_TIMER
            if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%kxmig_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weikxmig1






  !----------------------------------------------------------------  




!---------------------------------------------
!
!
!  Initialize storage for wavefields
!
!
subroutine init_wave_temps() 
 integer ::er
if( run%mem_save < 2) then
   tmp1=>tmp_on(:,:,:,:,1:run%o_nth)
   tmp2=>tmp_on(:,:,:,:,1+run%o_nth:2*run%o_nth)
else 
   if(run%mem_save==2) tmp1=>tmp_on(:,:,:,:,1:run%o_nth)
   allocate(tmps(down%wsep%n(1),down%wsep%n(2)))
   call add_mem(size(tmps))
   write(twave,"(a,a,a)") trim(run%local_path),"T_WAVE",run%tag_num
   call auxtmp(twave)
end if
end subroutine

!-------------------------------------------------
!
!
! Store the wavefield
!
!
subroutine store_w(wfld)
 complex :: wfld(:,:,:,:,:)
 integer :: ierr
    ierr= sseek(twave,0,0)
    ierr= srite(twave,wfld(:,:,:,:,1),8*size(wfld(:,:,:,:,1)))
end subroutine

subroutine store_wavefield(wfld,ith,iws,zero1)
 complex :: wfld(:,:,:,:,:)
 logical :: zero1
 integer :: ierr,iws,ith,i1,i2
  
if(run%mem_save>1) then
#ifdef SEP_TIMER
   if(run%m_ith==0 .and. ith==1)call start_timer_num(timer%diskio_t)
#endif
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
   if(run%mem_save>2) then
     tmps=cmplx(0.,0.)
      do i2=1,size(wfld,4)
        do i1=1,size(wfld,3)
           ierr=srite(twave,tmps,size(tmps)*8)
        end do
      end do
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1)call stop_timer_num(timer%diskio_t)
#endif
    else
      tmp1(:,:,:,:,ith)=0.; !if mem_save=1 one temporary
   end if
else if(zero1) then
  tmp2(:,:,:,:,ith) = wfld(:,:,:,:,iws)
  tmp1(:,:,:,:,ith) = 0.
else
  tmp1(:,:,:,:,ith) = wfld(:,:,:,:,iws)
  tmp2(:,:,:,:,ith) = 0.
end if
end subroutine

!------------------------------------------------------
!
!
!
!
subroutine grab_part_wvfld(wfld,ifk,ith,iws,izs)
integer :: ith,iws,ifk,ierr,izs
 complex :: wfld(:,:,:,:,:)

if(run%mem_save>1) then
#ifdef SEP_TIMER
   if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%diskio_t)
#endif
  ierr= sseek(twave,0,0)
  ierr= sreed(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%diskio_t)
#endif
  call partition_slow(wfld(:,:,:,:,iws),wfld(:,:,:,:,iws),ifk,izs,ith)
else if(run%mem_save==1) then
  call partition_slow(tmp2(:,:,:,:,ith),wfld(:,:,:,:,iws),ifk,izs,ith)
else
    if(slo%th(ith)%source) then
     where(slo%si_s(:,:,:,:,izs) == ifk)
       wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
     elsewhere
         wfld(:,:,:,:,iws) = 0.
     end where
    else
     where(slo%si_r(:,:,:,:,izs) == ifk)
       wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
     elsewhere
         wfld(:,:,:,:,iws) = 0.
     end where
    end if
end if
end subroutine
!------------------------------------------------------
!
!
!

subroutine add_wavefield(wfld,ith,iws)
complex :: wfld(:,:,:,:,:)
integer :: ith,iws,i1,i2,ierr

if(run%mem_save>2) then
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%diskio_t)
#endif
  ierr= sseek(twave,size(wfld)*8,0)
   do i2=1,size(wfld,4)
     do i1=1,size(wfld,3)
       ierr= sreed(twave,tmps,size(tmps)*8)
       wfld(:,:,i1,i2,iws)=wfld(:,:,i1,i2,iws)+tmps
     end do
   end do 
  ierr= sseek(twave,size(wfld)*8,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)));
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%diskio_t)
#endif
else
  tmp1(:,:,:,:,ith) = tmp1(:,:,:,:,ith) +  wfld(:,:,:,:,iws)
end if

end subroutine
subroutine grab_w(wfld)
complex :: wfld(:,:,:,:,:)
logical :: first
integer :: ith,iws,i1,i2,ierr
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%diskio_t)
#endif
  ierr= sseek(twave,0,0)
  ierr= sreed(twave,wfld(:,:,:,:,1),8*size(wfld(:,:,:,:,1)))
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%diskio_t)
#endif
end subroutine
subroutine grab_wavefield(wfld,iws,ith,first)
complex :: wfld(:,:,:,:,:)
logical :: first
integer :: ith,iws,i1,i2,ierr
if(run%mem_save>2) then
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%diskio_t)
#endif
  if(.not. first) then
     ierr= sseek(twave,0,0)
  else
     ierr= sseek(twave,size(wfld)*8,0)
  end if
  ierr= sreed(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%diskio_t)
#endif
else if(first) then 
  wfld(:,:,:,:,iws) = tmp1(:,:,:,:,ith)
else
  wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
end if
end subroutine

subroutine store_part_wavefield(wfld,ifk,ith,iws,izs)
 complex wfld(:,:,:,:,:)
  integer :: ifk,ith,iws,ierr,izs,i1,i2,i3,i4
if(run%mem_save==1) then
  call partition_slow(wfld(:,:,:,:,iws),tmp2(:,:,:,:,ith),ifk,izs,ith)
else if(run%mem_save>1) then
  call partition_slow(wfld(:,:,:,:,iws),wfld(:,:,:,:,iws),ifk,izs,ith)
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%diskio_t)
#endif
  ierr= sseek(twave,0,0)
   do i2=1,size(wfld,4)
     do i1=1,size(wfld,3)
       ierr= sreed(twave,tmps,size(tmps)*8)
       wfld(:,:,i1,i2,iws)=wfld(:,:,i1,i2,iws)+tmps
     end do
   end do 
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
#ifdef SEP_TIMER
  if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%diskio_t)
#endif
else
  do i4=1,size(wfld,4)
  do i3=1,size(wfld,3)
  do i2=1,size(wfld,2)
  do i1=1,size(wfld,1)
    if(slo%th(ith)%si(i1,i2,i3,i4,izs)==ifk) then
       tmp2(i1,i2,i3,i4,ith)=tmp2(i1,i2,i3,i4,ith)+wfld(i1,i2,i3,i4,iws)
     end if
  end do
  end do
  end do
  end do
end if
end subroutine
end module
