module wei_obj_W
  use wei_util
  use wei_down
  use wei_run
  use wei_var
  use wei_type
  use wei_obj_D
  !---------------
  implicit none

  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  type(wewfld), public :: W
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!

contains
  !----------------------------------------------------------------
  subroutine weibwfld_init(X,tag)
    type(wewfld),       intent(inout) :: X
    character(len=128), intent(in)    :: tag
#ifdef DEBUG
    character(len=128), parameter :: unit='weibwfld_init'
    if(run%debug) call in(unit)
#endif DEBUG


    select case(run%operator(1:5))
    case('.Rw.D')

       X%tag = tag
       call wavefield_param_init(W)

    case('dSwdR')

       X%tag = tag
       if(isfile(X%tag)) then
          call wavefield_history_init(X)
       else
          call seperr('W /= file')
       end if
    end select
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif DEBUG

  end subroutine weibwfld_init
  !----------------------------------------------------------------
  subroutine wavefield_history_init(X)
    type(wewfld), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_history_init'
    if(run%debug) call in(unit)
#endif

    call auxinout(X%tag)
    X%from = "history"

    call seperr("wavefield history init unimplemented")

    call wavefield_setup(X)
#ifdef DEBUG
    if(run%debug) call wavefield_report(X)
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_history_init
  !----------------------------------------------------------------
  subroutine wavefield_param_init(X)
    type(wewfld), intent(inout) :: X
    integer i
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_param_init'
    if(run%debug) call in(unit)
#endif

    call auxscr(X%tag)
    X%from = "parameter"

    do i=1,4
     call sep_put_data_axis_par(X%tag,i,down%wsep%n(i),down%wsep%o(i),down%wsep%d(i),down%wsep%label(i))
    end do
     call sep_put_data_axis_par(X%tag,5,down%dsep%n(5),down%dsep%o(5),down%dsep%d(5),down%rsep%label(5))
     call sep_put_data_axis_par(X%tag,6,down%rsep%n(5),down%rsep%o(5),down%rsep%d(5),down%rsep%label(5))
    call to_history("esize",8,X%tag)

    call wavefield_setup(X)
    call wavefield_assign(X,cmplx(0.))
#ifdef DEBUG
    if(run%debug) call wavefield_report(X)
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_param_init
  !----------------------------------------------------------------
  subroutine wavefield_setup(X)
    type(wewfld), intent(inout) :: X
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_setup'
    if(run%debug) call in(unit)
#endif

    !! DEPTH blocking
    if( (1.0*down%rsep%n(5))/down%nzs /= down%rsep%n(5)/down%nzs ) then
       call wavefield_report(W)
       call seperr('W.az.n is not a multiple of W.nzs')
    end if

    !! FREQUENCY blocking
    if( (1.0*down%dsep%n(5))/down%nws /= down%dsep%n(5)/down%nws ) then
       call wavefield_report(W)
       call seperr('W.aw.n is not a multiple of W.nws')
    end if

    down%nwb = down%dsep%n(5) / down%nws
    down%nzb = down%rsep%n(5) / down%nzs
    X%bites = 8* product(down%wsep%n(1:4))

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_setup
  !----------------------------------------------------------------
  subroutine wavefield_assign(X,v)
    type(wewfld), intent(inout) :: X
    complex, intent(in) :: v
    integer :: izb, izs
    integer :: iwb, iws
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_assign'
    if(run%debug) call in(unit)
#endif

!    bW%bin=v
    do izb=1,down%nzb
       do izs=1,down%nzs
          do iwb=1,down%nwb
             do iws=1,down%nws
!                call wavefield_write(X,bW%bin,izb,izs,iwb,iws)
             end do
          end do
       end do
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_assign
  !----------------------------------------------------------------  
  subroutine wavefield_report(X)
    type(wewfld), intent(in) :: X

    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Mx: ",down%wsep%n(1), down%wsep%o(1), down%wsep%d(1)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    My: ",down%wsep%n(2), down%wsep%o(2), down%wsep%d(2)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Hx: ",&
   down%wsep%n(3), down%wsep%o(3), down%wsep%d(3)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Hy: ",&
   down%wsep%n(4), down%wsep%o(4), down%wsep%d(4)
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     W: ",&
  down%dsep%n(5),  down%dsep%o(5),  down%dsep%d(5)," ->",down%nwb," blocks"
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     Z: ",&
down%rsep%n(5),  down%rsep%o(5),  down%rsep%d(5)," ->",down%nzb," blocks"
    write(0,*)'        ************************************************'
  end subroutine wavefield_report
  !--------------------------------------------------------------
  subroutine wavefield_read(X,data,izb,izs,iwb,iws)
    type(wewfld), intent(in) :: X
    complex, dimension(:,:,:,:,:), pointer :: data
    integer, intent(in) :: izb,izs,iwb,iws
    integer :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_read'
    if(run%depar) call in(unit)
#endif

    k=(izb-1)*down%nzs*down%nwb*down%nws &
    +         (izs-1)*down%nwb*down%nws &
    +                 (iwb-1)*down%nws &
    +                             iws

    stat=sseek_block(X%tag,k-1,X%bites,0)
    stat= sreed(X%tag, data(:,:,:,:,iws), X%bites)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine wavefield_read
  !----------------------------------------------------------------
  subroutine wavefield_write(X,data,izb,izs,iwb,iws)
    type(wewfld), intent(in) :: X
    complex, dimension(:,:,:,:,:), pointer :: data
    integer, intent(in) :: izb,izs,iwb,iws
    integer :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_write'
    if(run%depar) call in(unit)
#endif

    k=(izb-1)*down%nzs*down%nwb*down%nws &
    +         (izs-1)*down%nwb*down%nws &
    +                 (iwb-1)*down%nws &
    +                             iws

    stat=sseek_block(X%tag,k-1,X%bites,0)
    stat= srite(X%tag, data(:,:,:,:,iws), X%bites)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine wavefield_write
  !----------------------------------------------------------------
end module wei_obj_W
