module wei_sll
  !! WEI slowness operator
  !! for FK-FX downward continuation
  !! includes salt logic

  use wei_util
  use wei_obj_S
 use lloyd_mod
  !---------------
  implicit none

  real, dimension(:,:,:,:), pointer, private :: sxy
  real, dimension(:), pointer, private :: bound,ref_slow,s_slice
  logical, private, save  :: write_it,first_time

contains
!!$
!!$=head1 NAME
!!$
!!$ weisll_init - Initialize multiple  slowness  lloyd operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weisll_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize multiple slowness operator lloyd
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weisloN_init> L<weislo1> L<weislsN>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$

  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weisll_init() result(st)
    integer :: er,i
#ifdef DEBUG
    character(len=128), parameter :: unit='weislsN_init'
    if(run%debug) call in(unit)
#endif


    allocate(bound(down%nref),ref_slow(down%nref))
    allocate(s_slice(size(S%s_s%bin,1)*size(S%s_s%bin,2)))
    
   if(run%mem_save<1) then
      allocate(sxy(down%wsep%n(1),down%wsep%n(2),down%wsep%n(3),down%wsep%n(4)),stat=er)
      if(er/=0) call seperr('cant allocate sxy')
   end if

    call init_lloyd()
    write_it=.false.;
    if(1==getch("slow_choice","d",i).and. run%m_ith==0) then
       write_it=.true.
       if(run%mem_save>0) then
         call seperr("can not do memory save and slow_choice")
       end if
       do i=1,4
         call sep_put_data_axis_par("slow_choice",i,down%wsep%n(i),down%wsep%o(i),down%wsep%d(i),down%wsep%label(i))
       end do
       call sep_put_data_axis_par("slow_choice",5,down%rsep%n(5),down%rsep%o(5),down%wsep%d(5),"Axis")
     end if



#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weisll_init
!!$
!!$=head1 NAME
!!$
!!$ weisll - Multiple  slowness operator lloyd vel/find
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weisll()
!!$
!!$=head1 DESCRIPTION
!!$
!!$  multiple slowness operator lloyd
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weisloN_init> L<weislo1> L<weislsN_init>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  !----------------------------------------------------------------  
  !! N reference slownesses
  integer function weisll() result(st)
    integer :: izs,ifk,i4,i3,i2,i1
    real    :: ds, so, sz
    real    :: sl,su
    real    :: smin,smax
    integer :: imx, imy
    real    :: ssalt_true
    real    :: ssalt_mean,ssalt_sum
    real    :: slocal
    integer :: ins,ndo,ierr,n

#ifdef DEBUG
    character(len=128), parameter :: unit='weislsN'
    if(run%debug) call in(unit)
#endif

    !! spread the slowness in its PBA    
#ifdef SEP_TIMER
  if(run%m_ith==0) call start_timer_num(timer%slow_t)
#endif

    call fill_big_pba(S)
    do izs=1,down%nzs
       call sellect_slow_comp(.true.,1)
       slo%th(1)%nrs(izs) = down%nref
       if(run%mem_save<1) sxy(:,:,:,:) =real(S%pba(:,:,:,:,izs))
       s_slice=reshape(S%s_s%bin(:,:,izs),(/size(S%s_s%bin(:,:,izs))/))
       n= lloyd_go(s_slice,ref_slow,down%nref,bound,run%m_ith==0)
       slo%th(1)%nrs(izs)= n
       call lloyd_setup(izs) 
       !NOW DO THE RECEIVER
       if(write_it) call write_slow_map()
       call sellect_slow_comp(.false.,1) !SET APPROPRIATE POINTERS
       !FILL IN SLOWNESS FIELD IF REL
       if(run%mem_save<1) sxy(:,:,:,:) =aimag(S%pba(:,:,:,:,izs))
       !RERUN LLOYD IF DIFFERENT
       if(down%conv_wave) then
          s_slice=reshape(S%s_g%bin(:,:,izs),(/size(S%s_g%bin(:,:,izs))/))
          slo%th(1)%nrs(izs)= &
           lloyd_go(s_slice,ref_slow,down%nref,bound,run%m_ith==0)
        else 
          slo%th(1)%nrs(izs)= n
        end if 
       call lloyd_setup(izs) 
    end do
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
#ifdef SEP_TIMER
  if(run%m_ith==0) call stop_timer_num(timer%slow_t)
#endif
    st=OK
  end function weisll

  !----------------------------------------------------------------
   subroutine write_slow_map()
     integer izs,i1,i2,i3,i4
         do  izs=1,size(slo%th(1)%si,5)
         do  i4=1,size(sxy,4)
         do  i3=1,size(sxy,3)
         do  i2=1,size(sxy,2)
         do  i1=1,size(sxy,1)
           sxy(i1,i2,i3,i4)=slo%sz_s(izs,slo%th(1)%si(i1,i2,i3,i4,izs))
         end do;end do;end do;end do
         if(4*size(sxy)/=srite("slow_choice",sxy,size(sxy)*4)) then
           call seperr("trouble writing out slow_choice")
         end if
       end do

   end subroutine
     

subroutine lloyd_setup(izs)
  integer :: i1,i2,i3,i4,izs,ilast,ic,nc
  real :: bound2(0:down%nref)


      bound2(0)=0.
      do i1=1,down%nref
       bound2(i1)=bound(i1)
      end do


       slo%th(1)%sz(izs,1:size(ref_slow)) =ref_slow
       slo%th(1)%bound(izs,1:size(bound))= bound
     !REMOVE
!       slo%th(1)%sz(izs,1:size(ref_slow)) =maxval(sxy)
!       slo%th(1)%bound(izs,1:size(bound))= 99999.
     !END REMOVE
    if(down%maxslow) then !GSP requires maximum velocity/minimum slowness
       do i1=1,slo%th(1)%nrs(izs)-1
         slo%th(1)%sz(izs,i1)=bound(i1)
       end do
       slo%th(1)%bound(izs,slo%th(1)%nrs(izs))=99999.
       slo%th(1)%sz(izs,slo%th(1)%nrs(izs))=maxval(sxy)
    end if


   if(run%mem_save<1)  then

      if(slo%th(1)%nrs(izs)==1) then; slo%th(1)%si(:,:,:,:,izs)=1;
      else if(slo%th(1)%nrs(izs)==2) then
       do  i4=1,size(sxy,4);do  i3=1,size(sxy,3)
       do  i2=1,size(sxy,2);do  i1=1,size(sxy,1)
         if(sxy(i1,i2,i3,i4) < bound(1)) then 
          slo%th(1)%si(i1,i2,i3,i4,izs) =1
         else
          slo%th(1)%si(i1,i2,i3,i4,izs) =2
         end if
       end do;end do;end do;end do
      else

       slo%th(1)%si(1,1,1,1,izs)=1
       do while(sxy(1,1,1,1) > bound(slo%th(1)%si(1,1,1,1,izs)))
            slo%th(1)%si(1,1,1,1,izs)=slo%th(1)%si(1,1,1,1,izs)+1
       end do
       ilast=slo%th(1)%si(1,1,1,1,izs)
       ic=0
       do  i4=1,size(sxy,4)
       do  i3=1,size(sxy,3)
       do  i2=1,size(sxy,2)
       do  i1=1,size(sxy,1)
         if(sxy(i1,i2,i3,i4) < bound2(ilast-1)  .or. &
           sxy(i1,i2,i3,i4) > bound2(ilast)) then
             ic=ic+1
             ilast=1
             do while(sxy(i1,i2,i3,i4) > bound(ilast))
               ilast=ilast+1
             end do
         end if 
         slo%th(1)%si(i1,i2,i3,i4,izs)=ilast
       end do;end do;end do;end do
     end if
     end if


end subroutine

       
end module wei_sll




