module wei_wem
  !! Wave-Equation Migration
  !! full 3-D prestack

  use wei_fk_share_mod
  use wei_util
  use wei_down
  use wei_external
  use wei_var
  use wei_run
  use wei_fft
  use wei_obj_S
  use wei_kmap
  !---------------
  implicit none

  real,                           private :: dkxko
  integer,                        private :: nkxko
  real,                           private ::  kxko,rdkzko,idkzko

  real   , dimension(:), pointer, private ::  exp_kzko,ikzko,ckzko
  complex, dimension(:), pointer, private ::  kzko, tkzko
  complex, pointer, private :: w_extrap_1(:,:,:,:,:),w_extrap_2(:,:,:,:,:)

contains



  !----------------------------------------------------------------
!!$=head1 NAME
!!$
!!$weiwem_init -initialize  wave equation migration 
!!$
!!$=head1 SYNOPSIS
!!$
!!$ierr= weiwem_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation migration operator
!!$
!!$Grabs kxko=2. and nkxko=2001 from par
!!$
!!$=head1 LIBRARY
!!$
!!$sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weiwem> L<weiwem_init>
!!$
!!$=cut
!!$
  integer function weiwem_init() result(st)
    integer :: er
    integer ::i,imax
    real :: hxmax,hymax
#ifdef DEBUG
    character(len=128), parameter :: unit='weiwem_init'
    if(run%debug) call in(unit)
#endif

    !! init DSR
    call from_param("nkxko",nkxko,2001)
    call from_param( "kxko", kxko,2.)
    call dsr_init(kxko,nkxko,kzko,tkzko,exp_kzko,ikzko,dkxko,nkxko,idkzko,rdkzko)
!    call wemdsr_init(kxko,nkxko)
   


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weiwem_init




  integer function weiwem_gsp_init() result(st)
    integer :: er
    integer ::i,imax,n
    real :: hxmax,hymax
#ifdef DEBUG
    character(len=128), parameter :: unit='weiwem_init'
    if(run%debug) call in(unit)
#endif

    !! init DSR
    call from_param("nkxko",nkxko,2001)
    call from_param( "kxko", kxko,2.)
    call dsr_init(kxko,nkxko,kzko,tkzko,exp_kzko,ikzko,dkxko,nkxko,idkzko,rdkzko)
!    call wemdsr_init(kxko,nkxko)

      n=size(tmp_on,5)
     if(run%fk_extra .ne. 2) call seperr("expecting  fk_extra ==2 for gsp")
     w_extrap_1=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
     n=n-run%o_nth
     w_extrap_2=>tmp_on(:,:,:,:,n-run%o_nth+1:n)
     OK=weiwem_init()
     down%maxslow=.true.


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weiwem_gsp_init




!!$=head1 NAME
!!$
!!$weiwem - one depth step of  wave equation migration
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weiwem(iws,izs,ifk,ith,wfld)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item iws - integer
!!$
!!$  current frequency slice
!!$
!!$=item izs - integer
!!$
!!$  current depth slice
!!$
!!$=item ifk - integer
!!$
!!$  current fk operation
!!$
!!$=item ith - integer
!!$
!!$  current thread
!!$
!!$=item wfld - complex(:,:,:,:,:)
!!$
!!$  wavefield
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave Equation Migration operator
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weiwem_init> L<weicam>
!!$
!!$=cut
  !----------------------------------------------------------------
  integer function weiwem(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) ::iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    :: w, kos,kor,ko
    integer :: imx,imy,ihx,ihy 
    integer :: is,ir,i,iz,sign
    complex :: ikz
    real,pointer :: kx(:,:),ky(:,:)
    real    :: kss,krr,k,slow_0,fract
#ifdef DEBUG
    character(len=128), parameter :: unit='weiwem'
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call start_timer_num(timer%fk_t)
#endif

!    w  = down%bw%o + (iws-1)*down%bw%d
    w=down%w_local(iws)
   call calc_comp_depend_vars(w,slow_0,ko,kx,ky,sign,slo%th(ith)%source,izs,ifk)
    call calc_sign_depend_vars(fract,tkzko,var%dzb(izs),kzko,ko)

    do ihy=1,size(wfld,4)
      do ihx=1,size(wfld,3)
          do imy=1,down%wsep%n(2)
             do imx=1,down%wsep%n(1)
                k = sqrt(kx(imx,ihx)**2 + ky(imy,ihy)**2)
                i =max(1, min(int(1 + k/ko / dkxko) , nkxko))

                ikz= ko*tkzko(i) 

                !! kz(w) = SSR(kos,|ks|) 
                wfld(imx,imy,ihx,ihy,iws) = &
                wfld(imx,imy,ihx,ihy,iws) * &
                cexp( ikz * var%dzb(izs))

             end do
          end do
       end do

    end do

#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%fk_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weiwem
  !----------------------------------------------------------------
  integer function weiwem_gsp(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) ::iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    :: w, kos,kor,ko,ksq
    integer :: imx,imy,ihx,ihy 
    integer :: is,ir,i,iz,sign
    complex :: ikz
    real,pointer :: kx(:,:),ky(:,:)
    real    :: kss,krr,k,cract,slow_0,gamma,inv_gamma,myr,myi,kz,iko2,dz
    real    :: r_inv_norm_ko,i_inv_norm_ko,fract,j,inv_slow_0,inv_slow_0_cb
    complex ::  c_shift,c_shift1,ra,c_etak1,c_etak2,s
#ifdef DEBUG
    character(len=128), parameter :: unit='weiwem'
    if(run%depar) call in(unit)
#endif
#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1 ) call start_timer_num(timer%fk_t)
#endif


!    w  = down%bw%o + (iws-1)*down%bw%d
    w=down%w_local(iws)

    call weiift(ith,izs,ith,w_extrap_1)
!    call weiift(ith,izs,ith,w_extrap_2)
   
    !FIGURE OUT WHAT SOURCE WE ARE USING
   call calc_comp_depend_vars(w,slow_0,ko,kx,ky,sign,slo%th(ith)%source,izs,ifk)
    call calc_sign_depend_vars(fract,tkzko,var%dzb(izs),kzko,ko)


    cract=slow_0
    iko2=1./ko**2
    r_inv_norm_ko=1./(rdkzko)
    i_inv_norm_ko=1./(idkzko)
    inv_slow_0=1./slow_0
    inv_slow_0_cb=1./slow_0**3

    
   ! First GSP stuff
  ! Constants for spatial conrrection terms
  ! as given by the generalized-screen expansion
  s=cmplx(0,-w)
  dz=var%dzb(izs)


  w_extrap_1(:,:,:,:,ith)=s/2.*dz*w_extrap_1(:,:,:,:,ith)
!  w_extrap_2(:,:,:,:,ith)=-s/8.*dz*w_extrap_2(:,:,:,:,ith)

     

    


if(ko >0.) then
  do ihy=1,size(wfld,4)
   do ihx=1,size(wfld,3)
    do imy=1,down%wsep%n(2)
      do imx=1,down%wsep%n(1)
        if (abs(wfld(imx,imy,ihx,ihy,iws))>0.0) then
           ksq = kx(imx,ihx)**2 + ky(imy,ihy)**2
           k=iko2*ksq
           if (k<1.) then
             i= min(nkxko,max(1, int(1 + k*i_inv_norm_ko)))
             gamma=(ikzko(i)*cract )
               inv_gamma=1./max(gamma,.000001)
               j=fract*ikzko(i)
               c_shift=cmplx(cos(j),sin(j))
               c_etak1=inv_gamma-inv_slow_0
               c_etak2=(inv_gamma**3-inv_slow_0_cb)
             if(gamma < .01 ) then
               c_etak1=0.
               c_etak2=0.
             end if
           else
               i=max(1, min(nkxko,int(1 + (k-1.)*r_inv_norm_ko)))
               inv_gamma=1.0/max(.000001,(cract*exp_kzko(i)))
               j=fract*exp_kzko(i)
               c_shift=cmplx(exp(j), 0.0)
               c_etak1=cmplx(inv_gamma, -inv_slow_0)
               c_etak2=cmplx(inv_gamma**3,-inv_slow_0_cb)
             if(gamma < .01 ) then
               c_etak1=0.
               c_etak2=0.
             end if
           end if


!!BEGIN ORDER 1
                 ra = (w_extrap_1(imx,imy,ihx,ihy,ith)*c_etak1) &
                      / wfld(imx,imy,ihx,ihy,iws)
!END ORDER 1
!BEGIN ORDER 2
!                ra = ((w_extrap_1(imx,imy,ihx,ihy,ith)*c_etak1) &
!                      + (w_extrap_2(imx,imy,ihx,ihy,ith)*c_etak2)) &
!                      / wfld(imx,imy,ihx,ihy,iws)
!END ORDER2

                 myr=real(ra)
                 myi=aimag(ra)
                 ra=1+myr/cmplx(1,myi)
                 c_shift1=cexp(cmplx(0.0,myi))*ra/cabs(ra)
                 wfld(imx,imy,ihx,ihy,iws) = &
                 wfld(imx,imy,ihx,ihy,iws)*c_shift*c_shift1!*c_shift2
              else
                 wfld(imx,imy,ihx,ihy,iws) = cmplx(0.0,0.0)
              end if

              ! GSP stuff finishes here
           end do
        end do
     end do
   end do
  end if


#ifdef SEP_TIMER
             if(run%m_ith==0 .and. ith==1) call stop_timer_num(timer%fk_t)
#endif
#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weiwem_gsp

end module wei_wem
